package jcs.ast;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import jcs.ast.gen.*;
import jcs.ast.gen.JCS.SortNd;
import jcs.checker.AbstractMsg;
import jcs.checker.Destination;
import jcs.checker.GPDA;
import jcs.checker.Guard;
import jcs.checker.StateTransformer;
import jcs.types.Environment;

public class Transformer {
	BoxSpecificationNd boxspec;
	
	Set<Destination> incomingCallNodes = new HashSet<Destination>();
	Set<Destination> incomingMethodDeclarationNodes = new HashSet<Destination>();
	
	HashMap<Integer, SortNd> labelMap = new HashMap<Integer, SortNd>();
	HashMap<SortNd, Integer> labelRevMap = new HashMap<SortNd, Integer>();
	
	Environment env = new Environment();
	StateTransformer st = new StateTransformer(env);
	
	public Transformer(BoxSpecificationNd b) {
		boxspec = b;
		env.buildEnvironment(boxspec.term());
		new IncomingNodeSearch();
	}
	
	public GPDA buildAutomaton() {
		GPDA automaton = new GPDA();
		
		//initial transitions
		automaton.addTransitions(0, incomingMethodDeclarationNodes);
		
		//transitions from incoming method
		for (Destination d : incomingMethodDeclarationNodes) {
			SortNd node = labelMap.get(d.getLabel());
			automaton.addTransitions(d.getLabel(), next(node, node, new Guard()));
		}
		
		//transitions from outgoing method
		for (Destination d : incomingMethodDeclarationNodes) {
			HashSet<Destination> result = new HashSet<Destination>();
    		result.addAll(incomingCallNodes);
    		result.addAll(incomingMethodDeclarationNodes);
    		automaton.addTransitions(d.getLabel() + 1, result);
		}
		
		//transitions from incoming calls
		for (Destination d : incomingCallNodes) {
			SortNd node = labelMap.get(d.getLabel());
			automaton.addTransitions(d.getLabel(), next(node, null, new Guard()));
		}
		
		//transitions from outgoing calls
		for (Destination d : incomingCallNodes) {
			HashSet<Destination> result = new HashSet<Destination>();
    		result.addAll(incomingCallNodes);
    		result.addAll(incomingMethodDeclarationNodes);
    		automaton.addTransitions(d.getLabel() + 1, result);
		}
		
		return automaton;
	}
	
    private Set<Destination> next(SortNd node, SortNd context, Guard g) {
		if (node instanceof CallNd) {
			return next((CallNd) node, context, g);
		}
		if (node instanceof LoopNd) {
			return next((LoopNd) node, context, g);
		}
		if (node instanceof OrNd) {
			return next((OrNd) node, context, g);
		}
		if (node instanceof BlockNd) {
			return next((BlockNd) node, context, g);
		}
		if (node instanceof EmptyStatementNd) {
			return next((EmptyStatementNd) node, context, g);
		}
		if (node instanceof AssumeNd) {
			return next((AssumeNd) node, context, g);
		}
		if (node instanceof AssertNd) {
			return next((AssertNd) node, context, g);
		}
		if (node instanceof MethodDeclarationNd) {
			return next((MethodDeclarationNd) node, context, g);
		}
		if (node instanceof BreakNd) {
			return next((BreakNd) node, context, g);
		}
		throw new RuntimeException("match missing");
	}
	
	
	
    private Set<Destination> next(CallNd call, SortNd context, Guard g) {
    	if (call == context) {
    		HashSet<Destination> result = new HashSet<Destination>();
    		int label = labelRevMap.get(call) + 1;
    		result.add(new Destination(new AbstractMsg(env, call, false), label, g));
        	return result;
    	}
    	if (context == null) {
    		return next(call.parent(), call, g);
    	}
    	throw new RuntimeException("wrong tree traversal");
	}
	
    private Set<Destination> next(LoopNd loop, SortNd context, Guard g) {
    	if (context instanceof BreakNd) {
    		return next(loop.parent(), loop, g);
    	}
		if (context == loop || context == loop.statement()) {
			return next(loop.statement(), loop.statement(), g);
		}
		throw new RuntimeException("wrong tree traversal");
	}
    
    private Set<Destination> next(BreakNd breaknd, SortNd context, Guard g) {
		return next(breaknd.parent(), breaknd, g);
	}
	
    private Set<Destination> next(OrNd or, SortNd context, Guard g) {
    	if (context instanceof BreakNd) {
    		return next(or.parent(), context, g);
    	}
    	if (context == or) {
			Set<Destination> temp = new HashSet<Destination>();
			for (StatementNd s : or) {
				temp.addAll(next(s, s, g));
			}
			return temp;
		}
		if (context.parent() == or) {
			return next(or.parent(), or, g);
		}
		throw new RuntimeException("wrong tree traversal");
	}
	
    private Set<Destination> next(BlockNd block, SortNd context, Guard g) {
    	if (context instanceof BreakNd) {
    		return next(block.parent(), context, g);
    	}
		if (context == block) {
			if (block.size() == 0) {
				return next(block.parent(), block, g);
			} else {
				return next(block.first(), block.first(), g);
			}
		}
		if (context.parent() == block) {
			if (context == block.last()) {
				return next(block.parent(), block, g);
			} else {
				StatementNd nextNode = block.get(context.position() + 1);
				return next(nextNode, nextNode, g);
			}
		}
		throw new RuntimeException("wrong tree traversal");
	}
    
    private Set<Destination> next(EmptyStatementNd es, SortNd context, Guard g) {
    	if (es == context) {
    		return next(es.parent(), es, g);
    	}
    	throw new RuntimeException("wrong tree traversal");
    }
    
    private Set<Destination> next(AssumeNd es, SortNd context, Guard g) {
    	if (es == context) {
    		return next(es.parent(), es, g);
    	}
    	throw new RuntimeException("wrong tree traversal");
    }
    
    private Set<Destination> next(AssertNd es, SortNd context, Guard g) {
    	if (es == context) {
    		Guard g1 = st.toGuard(es.expr().termExpression());
    		return next(es.parent(), es, g.and(g1));
    	}
    	throw new RuntimeException("wrong tree traversal");
    }
    
    private Set<Destination> next(MethodDeclarationNd es, SortNd context, Guard g) {
    	if (es == context) {
    		return next(es.block(), es.block(), g);
    	}
    	if (es.block() == context) {
    		HashSet<Destination> result = new HashSet<Destination>();
    		int label = labelRevMap.get(es) + 1;
    		result.add(new Destination(new AbstractMsg(env, es, false), label, g));
        	return result;
    	}
    	throw new RuntimeException("wrong tree traversal");
    }
    
	private class IncomingNodeSearch {
		int currval = 0;
		
		int incr() {
			return currval = currval + 2;
		}
		
		IncomingNodeSearch() {
			getAllCallNodes();
		}
		
		void getAllCallNodes() {
			for (MatchDeclarationNd match : boxspec.box().matches()) {
				for (MethodDeclarationNd method : match.methods()) {
					int label = incr();
					List<AssumeNd> assms = new AssumeNodeSearch(method).found;
					Guard guard = new Guard();
					for (AssumeNd assm : assms) {
						Guard ng = st.toGuard(assm.expr().termExpression());
						guard = guard.and(ng);
					}
					incomingMethodDeclarationNodes.add(new Destination(new AbstractMsg(env, method, true), label, guard));
					labelMap.put(label, method);
					labelRevMap.put(method, label);
					
					search(method.block());
				}
			}
		}
		
		void search(StatementNd s) {
			if (s instanceof BlockNd) {
				for (StatementNd n : s.<BlockNd>cast()) {
					search(n);
				}
			}
			if (s instanceof OrNd) {
				for (StatementNd n : s.<OrNd>cast()) {
					search(n);
				}
			}
			if (s instanceof LoopNd) {
				search(s.<LoopNd>cast().statement());
			}
			if (s instanceof CallNd) {
				CallNd call = s.<CallNd>cast();
				//callnodes.add(s.<CallNd>cast());
				int label = incr();
				List<AssumeNd> assms = new AssumeNodeSearch(s).found;
				Guard guard = new Guard();
				for (AssumeNd assm : assms) {
					Guard ng = st.toGuard(assm.expr().termExpression());
					guard = guard.and(ng);
				}
				incomingCallNodes.add(new Destination(new AbstractMsg(env, call, true), label, guard));
				labelMap.put(label, s);
				labelRevMap.put(s, label);
			}
		}
	}
	
	
	private class AssumeNodeSearch {
		class StopException extends Exception {

		}

		List<AssumeNd> found = new ArrayList<AssumeNd>();
		
		AssumeNodeSearch(SortNd s) {
			try {
			  beginSearch(s);
			} catch (StopException e) {}
			//System.out.println("Found " + found.size());
		}
		
		void beginSearch(SortNd s) throws StopException {
			//either MethodDeclarationNd or CallNd
			if (s instanceof MethodDeclarationNd) {
				search(s.<MethodDeclarationNd>cast().block());
				return;
			}
			if (s instanceof CallNd) {
				SortNd parent = s.parent(); 
				if (parent instanceof BlockNd) {
					BlockNd pp = parent.<BlockNd>cast();
					for (int i = s.position() + 1; i < pp.size(); i++) {
						search(pp.get(i));
					}
					return;
				}
			}
			throw new StopException();
		}
		
		void search(StatementNd s) throws StopException {
			if (s instanceof BlockNd) {
				for (StatementNd n : s.<BlockNd>cast()) {
					search(n);
				}
			}
			if (s instanceof AssumeNd) {
				found.add(s.<AssumeNd>cast());
				return;
			}
			
			throw new StopException();
		}
	}
}
