package jcs.types;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import jcs.ast.gen.*;

public class Environment {
	private HashMap<jcs.ast.gen.Type, Type> map = new HashMap<jcs.ast.gen.Type, Type>();
	
	private Set<StateType> st = new HashSet<StateType>();
	
	public void buildEnvironment(BoxSpecification b) {
		buildStates(b.box().states());
		buildTypes(b.box().matches());
	}

	public StateType getStateType(String s) {
		for (StateType stype : st) {
			if (stype.hasName(s)) {
				return stype;
			}
		}
		throw new IllegalArgumentException("Type not found " + s);
	}
	
	public Type getType(jcs.ast.gen.Type type) {
		Type t = map.get(type);
		if (t == null) {
			throw new IllegalArgumentException("Type not found " + type);
		}
		return t;
	}
	
	
	
	private void buildTypes(MatchDeclarationList matches) {
		for (MatchDeclaration match : matches) {
			for (jcs.ast.gen.Type type : match.types()) {
				put(type);
			}
			for (MethodDeclaration method : match.methods()) {
				put(method.type());
				for (Parameter p : method.params()) {
					put(p.type());
				}
				search(method.block());
			}
		}
	}
	
	void search(Statement s) {
		if (s instanceof Block) {
			for (Statement n : s.<Block>cast()) {
				search(n);
			}
		}
		if (s instanceof Or) {
			for (Statement n : s.<Or>cast()) {
				search(n);
			}
		}
		if (s instanceof Loop) {
			search(s.<Loop>cast().statement());
		}
		if (s instanceof Call) {
			Call call = s.<Call>cast();
			put(call.type());
			for (jcs.ast.gen.Parameter p : call.params()) {
				put(p.type());
			}
		}
	}

	private void buildStates(StateDeclarationList states) {
		for (StateDeclaration s : states) {
			Set<String> values = new HashSet<String>();
			for (Identifier id : s.idents()) {
				values.add(id.child0());
			}
			st.add(new StateType(s.ident().child0(), values));
		}
	}
	
	private Type buildType(jcs.ast.gen.Type type) {
		if (type instanceof jcs.ast.gen.WildcardOne) {
			return WildcardOneType.instance();
		}
		if (type instanceof jcs.ast.gen.WildcardAny) {
			return WildcardAnyType.instance();
		}
		if (type instanceof jcs.ast.gen.Identifier) {
			// TODO: consider imports here
			String s = type.<jcs.ast.gen.Identifier>cast().child0();
			try {
				return new ConcreteType(Class.forName(s));
			} catch (ClassNotFoundException e) {
				throw new IllegalArgumentException("No valid type: " + s);
			}
		}
		if (type instanceof jcs.ast.gen.IdentifierList) {
			String s = "";
			for (jcs.ast.gen.Identifier id : type.<jcs.ast.gen.IdentifierList>cast()) {
				String s2 = id.child0();
				s = s + s2 + ".";
			}
			s = s.substring(0, s.length() - 1);
			try {
				return new ConcreteType(Class.forName(s));
			} catch (ClassNotFoundException e) {
				//TODO: try out all package prefixes
				try {
					return new ConcreteType(Class.forName("java.lang." + s));
				} catch (ClassNotFoundException e2) {
				    throw new IllegalArgumentException("No valid type: " + s);
				}
			}
		}
		if (type instanceof jcs.ast.gen.IntersectionType) {
			Set<Type> tt = new HashSet<Type>();
			for (jcs.ast.gen.Type stype : type.<jcs.ast.gen.IntersectionType>cast()) {
				tt.add(buildType(stype));
			}
			return new IntersectionType(tt);
		}
		if (type instanceof jcs.ast.gen.BasicType) {
			String s = type.<jcs.ast.gen.BasicType>cast().child0();
			if (s.equals("boolean"))
				return new ConcreteType(Boolean.TYPE);
			if (s.equals("byte"))
				return new ConcreteType(Byte.TYPE);
			if (s.equals("char"))
				return new ConcreteType(Character.TYPE);
			if (s.equals("short"))
				return new ConcreteType(Short.TYPE);
			if (s.equals("int"))
				return new ConcreteType(Integer.TYPE);
			if (s.equals("long"))
				return new ConcreteType(Long.TYPE);
			if (s.equals("float"))
				return new ConcreteType(Float.TYPE);
			if (s.equals("double"))
				return new ConcreteType(Double.TYPE);
			if (s.equals("void"))
				return new ConcreteType(Void.TYPE);
		}
		throw new IllegalArgumentException("No valid type");
	}

	private void put(jcs.ast.gen.Type key) {
		if (!map.containsKey(key)) {
			map.put(key, buildType(key));
			//System.out.println("Putting " + key + "  :  " + buildType(key));
		}
	}
}
