package xtc.lang.jeannie;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import xtc.tree.GNode;

/**
 * The Blink terminal debugger. This debugger internally uses jdb and gdb to
 * support the Jeannie source level debugging.
 * 
 * @author Byeongcheol (BK) Lee
 */
public class Debugger {

  /**
   * The debugger control status.
   */
  public enum DebugerControlStatus {
    NONE, JDB, GDB, JDB_IN_GDB, GDB_IN_JDB,
  }

  /**
   * The current state of jdb process.
   */
  private enum ProcessState {
    NOT_CREATED, LIVE, DEAD,
  }

  /**
   * Prints a usage message to the user, and terminate the Blink debugger with
   * an error code(-1).
   * 
   * @param msg A message to the user to show what is wrong in the command line.
   */
  private static void usage(String msg) {
    StringBuffer buf = new StringBuffer();
    if (msg != null && msg.length() > 0) {
      buf.append(msg).append("\n\n");
    }
    buf.append("Usage: xtc.lang.jennie.Debugger [options] CLASS [arguments]\n");
    buf.append("Blink options include:\n");
    buf.append("\t-help\n");
    buf.append('\n');

    buf.append("options forwarded to jdb include:\n");
    buf.append("\t-sourcepath <directories separated by \":\">\n");
    buf.append("\t-attach <address>\n");
    buf.append("\t-listen <address>\n");
    buf.append("\t-listenany\n");
    buf.append("\t-launch\n");
    buf.append("\t-dbgtrace [flags]\n");
    buf.append("\t-tclient\n");
    buf.append("\t-tserver\n");
    buf.append('\n');

    buf.append("options forwarded to debuggee through jdb include:\n");
    buf.append("\t-v -verbose[:class|gc|jni]\n");
    buf.append("\t-D<name>=<value>  system property\n");
    buf.append("\t-classpath <directories separated by \":\">\n");
    buf.append("\t-X<option>\n");

    System.out.println(buf.toString());
    System.exit(-1);
  }

  /**
   * Print the help message for the Blink command.
   * 
   */
  void printCommandHelp() {
    final StringBuilder sb = new StringBuilder();
    sb.append("\n");
    sb.append("help               print help\n");
    sb.append("exit               exit the Blink debugger\n");
    sb.append("run                start the program run\n");
    sb.append("\n");
    sb.append("break [file:line]         add a break point e.g.) break Main.jni:9\n");
    sb.append("stop at <classid>:<line>  add a break point  e.g.) stop at Main:15\n");
    sb.append("watch <expr>              add a watch point \n");
    sb.append("watch [access/all] <class id>.<fieldname>  add a watch point\n");
    sb.append("info break                list break points\n");
    sb.append("info watch                list watch points\n");
    sb.append("delete [n]                delete a break/watch point with its id [n].\n");
    sb.append("\n");
    sb.append("where                     dump stack trace\n");
    sb.append("up n                      select n frames up\n");
    sb.append("down n                    select n frames down\n");
    sb.append("locals                    print local variables in selected frame\n");
    sb.append("\n");
    sb.append("continue                  coninue running.\n");
    sb.append("step                      execute until another line reached\n");
    sb.append("next                      execute the next line, including function calls\n");
    sb.append("\n");
    sb.append("print <jexpr>             print Jeannie expression\n");
    sb.append("list                      print source code.\n");
    out(sb.toString());
  }

  /**
   * The main method for the Blink debugger.
   * 
   * @param args The command line arguments.
   */
  public static void main(String[] args) {
    StringBuffer sbJDBOptions = new StringBuffer();
    StringBuffer sbGDBOPtions = new StringBuffer();
    String mainClass = null;
    StringBuffer sbMainOptions = new StringBuffer();

    // parse arguments
    for (int i = 0; i < args.length; i++) {
      String arg = args[i];
      // [options] CLASS
      if (mainClass == null) {
        if (arg.equals("-help")) {
          usage("");
        } else if (arg.equals("-sourcepath")) {
          if ((i + 1) >= args.length)
            usage("Please, specify path after -sourcepath.");
          String argPath = args[++i];
          sbJDBOptions.append(' ').append(arg).append(' ').append(argPath);
        } else if (arg.equals("-attach")) {
          if ((i + 1) >= args.length)
            usage("Please, specify <address> after -attach.");
          String argPath = args[++i];
          sbJDBOptions.append(' ').append(arg).append(' ').append(argPath);
        } else if (arg.equals("-dbgtrace")) {
          sbJDBOptions.append(' ').append(arg);
        } else if (arg.equals("-classpath")) {
          if ((i + 1) >= args.length)
            usage("Please, specify path after -classpath.");
          String argPath = args[++i];
          sbJDBOptions.append(' ').append(arg).append(' ').append(argPath);
        } else if (arg.matches("-D[^=]+=.*")) {
          sbJDBOptions.append(' ').append(arg);
        } else {
          mainClass = arg;
        }
      } else {
        // [arguments]
        sbMainOptions.append(' ').append(args);
      }
    }

    // check arguments
    if (mainClass == null) {
      usage("Please, specify the main CLASS name.");
    }

    // Now, run the Blink debugger.
    StringBuffer jdbArguments = new StringBuffer();
    jdbArguments.append(sbJDBOptions);
    jdbArguments.append(' ').append(mainClass);
    jdbArguments.append(' ').append(sbMainOptions);
    Debugger jeannieDebugger = new Debugger(jdbArguments.toString(),
        sbGDBOPtions.toString());
    try {
      jeannieDebugger.startDebugging();
      jeannieDebugger.runUserCommandLoop();
    } catch (Exception e) {
      e.printStackTrace();
    } finally {
      jeannieDebugger.ensureResourceRelease();
    }

  }

  /**
   * The option string to be forwarded to the jdb.
   */
  private final String jdbArguments;

  /**
   * The jdb slave process.
   */
  private SlaveProcessHandler jdbHandler;

  /**
   * The option string to be forwarded to the gdb.
   */
  private final String gdbOptions;

  /**
   * The gdb slave process.
   */
  private SlaveProcessHandler gdbHandler;

  /**
   * This flag is set to be true at the beginning. If any slave process is dead,
   * this flag is set to be false, indicating that the current debugging session
   * should terminate.
   */
  boolean allSlaveProcessLive = true;

  /**
   * The current debugger control status.
   */
  private DebugerControlStatus debugerControlStatus = DebugerControlStatus.NONE;

  /**
   * The user terminal input stream.
   */
  private final BufferedReader userInput;

  /**
   * The user terminal output stream.
   */
  private final BufferedOutputStream userOutput;

  /**
   * The user terminal error stream.
   */
  private final BufferedOutputStream userError;

  /**
   * This is a flag to recode a fact that the jdb and gdb are initialized for
   * j2c and c2j debugger switching.
   */
  private boolean debuggerSwitchingInitialized = false;

  /**
   * The actual implementation of the blink macro command.
   */
  final DebuggerCommand debuggerCommand;

  /**
   * The Blink command parser.
   */
  final DebuggerInterpreter interpreter;

  /**
   * @param jdbArguments The jdb command line options.
   * @param gdbOptions The gdb command line options.
   */
  private Debugger(final String jdbArguments, final String gdbOptions) {
    this.jdbArguments = jdbArguments;
    this.gdbOptions = gdbOptions;
    //prepares terminal I/O members
    this.userInput = new BufferedReader(new InputStreamReader(System.in));
    this.userOutput = new BufferedOutputStream(System.out);
    this.userError = new BufferedOutputStream(System.err);

    this.debuggerCommand = new DebuggerCommand(this);
    this.interpreter = new DebuggerInterpreter(this);
  }

  /**
   * Start debugging processes with jdb.
   */
  private void startDebugging() throws IOException {
    //jdb - attach the debugee
    String jdbCommand = "jdb " + jdbArguments;
    out("lanuching jdb : " + jdbCommand);
    jdbHandler = new SlaveProcessHandler("jdb", jdbCommand);
    jdbHandler.requestExpect("Initializing jdb ...");
    jdbHandler.start();
    jdbHandler.waitForExpect();
    jdbHandler.runAndWait("stop in DebugHelper.c2j\n",
        "Deferring breakpoint DebugHelper.c2j");
    jdbHandler.runAndWait("stop in java.lang.System.loadLibrary\n",
        "Deferring breakpoint java.lang.System.loadLibrary.");
    out("jdb is initialized.");
  }

  /**
   * Run the user command loop. Ask the user a command, and redirect the command
   * to the Blink macro command executer. This command loop may return if any
   * slave processes such as jdb, gdb and debuggee terminates or if the user
   * asks the termination with "exit" command.
   */
  private void runUserCommandLoop() {
    try {
      changeDebugStatus(DebugerControlStatus.JDB);
      sendUserMessage("\n");

      // The user command loop.
      for (String line = userInput.readLine(); line != null
          && allSlaveProcessLive; line = userInput.readLine()) {
        final String language;
        switch (debugerControlStatus) {
        case JDB_IN_GDB:
        case JDB:
          language = "Java";
          break;
        case GDB_IN_JDB:
        case GDB:
          language = "C";
          break;
        default:
          language = null;
          assert false;
          break;
        }
        final Object astOrMsg = Utilities.debuggerParseAndAnalyze(language,
            line, debuggerCommand.dbgExpr);
        if (astOrMsg instanceof GNode) {
          final GNode ast = (GNode) astOrMsg;
          if (ast.hasName("ExitCommand"))
            break;
          interpreter.dispatch(ast);
        } else {
          err((String) astOrMsg);
        }
      }
    } catch (IOException e) {
      System.err.println("error in the system input stream");
      e.printStackTrace();
    }
  }

  /**
   * Implement Blink "run" command.
   */
  void run() {
    try {
      assert ( getControlStatus() == DebugerControlStatus.JDB);
       runAndWaitJDB("run\n", "run");
       changeDebugStatus(Debugger.DebugerControlStatus.NONE);
    } catch (IOException e) {
       err("could not successfully run the application");
    }
  }

  /**
   * Implements the j2c macro command. This will activate the gdb assuming that
   * the jdb has control over the debuggee.
   */
  void j2n() {
    if (!IsDebuggerSwitchingInitialized()) {
      out("please run initj before j2c");
      return;
    }
    out("running j2c and switching gdb user interaction...");
    try {
      runJDBAndWaitGDB(
            "eval DebugHelper.j2c()\n",
            "Program received signal SIGTRAP, Trace/breakpoint trap.");
      changeDebugStatus(DebugerControlStatus.GDB_IN_JDB);
    } catch (IOException e) {
      err("failed in executing j2c\n");
      e.printStackTrace();
    }
  }

  /**
   * Implements c2j macro command. This will activate the jdb assuming that gdb
   * has control over the debugee.
   */
  void n2j() {
    if (!IsDebuggerSwitchingInitialized()) {
      out("please run initj before c2j");
      return;
    }
    out("running c2j and switching jdb user interaction...");
    try {
      runGDBAndWaitJDB(
          "call DebugHelper_c2j()\n", "Breakpoint hit:");
      changeDebugStatus(DebugerControlStatus.JDB_IN_GDB);
    } catch (IOException e) {
      err("failed in executing c2j\n");
      e.printStackTrace();
    }
  }

  /**
   * Implements jret macro command. If the current debugger nesting is jdb->gdb,
   * then this will resume the gdb break point, and the user input redirection
   * will be switched to the jdb. If the current debugger nesting is gdb->jdb,
   * this will resume the jdb break point, and the user input redirection will
   * be switched to the gdb.
   */
  void jret() {
    if (!IsDebuggerSwitchingInitialized()) {
      out("please run initj before jret");
      return;
    }
   switch(getControlStatus()) {
   case NONE:
   case JDB:
   case GDB:
     out("jret is for jdb and gdb nesting.");
     break;
   case JDB_IN_GDB:
     out("return to gdb");
     try {
       runJDBAndWaitGDB("cont\n", "= 1");
       changeDebugStatus(DebugerControlStatus.GDB);
     } catch(IOException e) {
      err("failed in resuming jdb control nested in the gdb"); 
     }
     break;
   case GDB_IN_JDB:
     out("returning to jdb");
     try {
       runGDBAndWaitJDB("continue\n", "DebugHelper.j2c\\(\\) = <void value>");
       changeDebugStatus(DebugerControlStatus.JDB);
     } catch(IOException e) {
       err("failed in resuming gdb control nested in the jdb");
     }
     break;
   default:
     break;
   } 
  }  

  /**
   * Implement Blink "continue/cont" command.
   * 
   */
   void cont() {
    switch (getControlStatus()) {
    case JDB:
      try {
        sendUserMessage("cont\n");
        changeDebugStatus(Debugger.DebugerControlStatus.NONE);
      } catch (IOException e) {
        err("failed in executing the continue");
      }
      break;
    case GDB:
      try {
        sendUserMessage("continue\n");
        changeDebugStatus(Debugger.DebugerControlStatus.NONE);
      } catch (IOException e) {
        err("failed in executing the continue");
      }
      break;
    case JDB_IN_GDB:
      out("\"continue\" is not allowed in jdb nested in the gdb.");
      out("use jret to return to the gdb");
      break;
    case GDB_IN_JDB:
      out("\"continue\" is not allowed in gdb nested in the jdb.");
      out("use jret to return to the jdb");
      break;
    }
  }

  /**
   * @return Wheather or not the jdb/gdb switching is ready.
   */
  boolean IsDebuggerSwitchingInitialized() {
    return debuggerSwitchingInitialized;
  }

  /**
   * Prepare jdb and gdb for j2c and c2j debugger switching.
   */
  void preparingDebuggerSwitching() {
    if (IsDebuggerSwitchingInitialized()) {
      out("debuggers were previously initialized for switching.");
      return;
    }

    // try to initialize the DebugHelper
    try {
      runAndWaitJDB("eval java.lang.Class.forName(\"DebugHelper\")\n",
          "java.lang.Class.forName\\(\"DebugHelper\"\\) = \"class DebugHelper\"");
      runAndWaitJDB("eval DebugHelper.init()\n",
          "DebugHelper.init\\(\\) = <void value>");
    } catch (IOException e) {
      err("failed in executing initj for jdb\n");
      return;
    }

    // get debugee process id through the DebugHelper Code.
    int debugeeProcessID;
    try {
      Matcher m = runAndWaitJDB("print DebugHelper.getProcessID()\n",
          "DebugHelper.getProcessID\\(\\) = (\\d+)");
      debugeeProcessID = Integer.parseInt(m.group(1));
    } catch (IOException e) {
      err("failed in getting debugged process id\n");
      return;
    }

    // gdb - attach the debugee
    try {
      String gdbCommand = "gdb -nw -quiet --pid " + debugeeProcessID
          + " -ex continue " + gdbOptions;
      gdbHandler = new SlaveProcessHandler("gdb", gdbCommand);
      gdbHandler.requestExpect("Continuing");
      gdbHandler.start();
      gdbHandler.waitForExpect();
      out("gdb is initialized.");
      sendUserMessage("\n"); // trigger 'jdb' prompt.
    } catch (IOException e) {
      err("failed in \n");
      return;
    }

    debuggerSwitchingInitialized = true;
  }

  /**
   * Send a message to the jdb , and wait until the jdb sends back an expected
   * message.
   * 
   * @param msg The message to send to the jdb.
   * @param expect The message to be expected from the jdb.
   * @return The matched result for the expected messsage.
   */
  Matcher runAndWaitJDB(String msg, String expect) throws IOException {
    return jdbHandler.runAndWait(msg, expect);
  }

  /**
   * Send a message to the jdb, and wait for a number of expected messages from
   * the jdb.
   * 
   * @param msg The message to send to jdb.
   * @param expects The list of expected message message.
   * @return The response from the jdb.
   */
  DebuggerResponse runAndWaitJDB(String msg, String[] expects)
      throws IOException {
    return jdbHandler.runAndWait(msg, expects);
  }

  /**
   * Send a message to the gdb process, and wait until the gdb sends back an
   * expected message.
   * 
   * @param msg The message to send.
   * @param expect The message to be expected from the gdb.
   */
  Matcher runAndWaitGDB(String msg, String expect) throws IOException {
    return gdbHandler.runAndWait(msg, expect);
  }

  /**
   * Send a message to the jdb process, and wait until the gdb sends back an
   * expected message.
   * 
   * @param msg The message to send.
   * @param expect The message to be expected from the slave process.
   */
  Matcher runJDBAndWaitGDB(String msg, String expect) throws IOException {
    return jdbHandler.runAndWait(msg, expect, gdbHandler);
  }

  /**
   * Send a message to the gdb process, and wait until the jdb sends back an
   * expected message.
   * 
   * @param msg The message to send.
   * @param expect The list of the messages expected from the jdb.
   */
  Matcher runGDBAndWaitJDB(String msg, String expect) throws IOException {
    return gdbHandler.runAndWait(msg, expect, jdbHandler);
  }

  /**
   * Print a message to the console.
   * 
   * @param msg The message to print.
   */
  void out(String msg) {
    try {
      userOutput.write(msg.getBytes());
      userOutput.flush();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  /**
   * Print an error message to the console.
   * 
   * @param msg The error message to print.
   */
  void err(String msg) {
    try {
      String msgAll = "\n[blink] error: " + msg + "\n";
      userError.write(msgAll.getBytes());
      userError.flush();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  /**
   * Switch to the new debugger state.
   * 
   * @param newStatus The requested new state.
   */
  synchronized void changeDebugStatus(DebugerControlStatus newStatus) {
    if (newStatus == debugerControlStatus) {
      return;
    }
    boolean success = false;
    switch (debugerControlStatus) {
    case NONE:
      success = newStatus == DebugerControlStatus.JDB
          || newStatus == DebugerControlStatus.GDB;
      break;
    case JDB:
      success = newStatus == DebugerControlStatus.NONE
          || newStatus == DebugerControlStatus.GDB_IN_JDB;
      break;
    case GDB:
      success = newStatus == DebugerControlStatus.NONE
          || newStatus == DebugerControlStatus.JDB_IN_GDB;
      break;
    case JDB_IN_GDB:
      success = newStatus == DebugerControlStatus.GDB;
      break;
    case GDB_IN_JDB:
      success = newStatus == DebugerControlStatus.JDB;
      break;
    }

    if (success) {
      debugerControlStatus = newStatus;
    } else {
      err("Not allowed transition:" + debugerControlStatus + "->" + newStatus);
      Thread.dumpStack();
    }
  }

  /**
   * @return The debugger control status.
   */
  DebugerControlStatus getControlStatus() {
    return debugerControlStatus;
  }

  /**
   * Show the current Blink status.
   */
  void printStatus() {
    out("control: " + debugerControlStatus);
  }

  /**
   * Send a message to the currently chose debugger.
   * 
   * @param msg The message.
   */
  synchronized void sendUserMessage(String msg) throws IOException {
    switch (debugerControlStatus) {
    case NONE:
      err("Not sure where to send this messsage:" + msg);
      break;
    case JDB:
    case JDB_IN_GDB:
      jdbHandler.sendMessage(msg);
      break;
    case GDB:
    case GDB_IN_JDB:
      gdbHandler.sendMessage(msg);
      break;
    default:
      err("Unknown debug control status.");
      break;
    }
  }

  /**
   * Notification of the death of the slave process.
   * 
   * @param slave The dead process handler.
   */
  private synchronized void onSlaveProcessDeath(SlaveProcessHandler slave) {
    if (slave == jdbHandler) {
      out("jdb finished...");
    } else if (slave == gdbHandler) {
      out("gdb finished...");
    }
    allSlaveProcessLive = false;
  }

  /**
   * Process the output message from the slave process. 
   * 
   * @param slave The handler of the slave process.
   * @param buf The byte buffer.
   * @param len The length of the data in the buffer.
   */
  private void onSlaveProcessOut(SlaveProcessHandler slave, byte[] buf, int len) {
    try {
      userOutput.write(buf, 0, len);
      userOutput.flush();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  /**
   * Process the error message form the slave process.
   * 
   * @param slave The handler of the slave process.
   * @param buf The byte buffer.
   * @param len The length of the data in the buffer.
   */
  private void onSlaveProceeErr(SlaveProcessHandler slave, byte[] buf, int len) {
    try {
      userError.write(buf, 0, len);
      userError.flush();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  /**
   * The process break point hit notification from the slave.
   * @param slave The slave process that gets the break point.
   */
  private synchronized void onSlaveProcessBreakPointHit(SlaveProcessHandler slave) {
    if (slave == jdbHandler) {
      changeDebugStatus(DebugerControlStatus.JDB);
    } else if (slave == gdbHandler) {
      changeDebugStatus(DebugerControlStatus.GDB);
    }
  }

  /**
   * The jdb hits the System.loadLibrary event.
   * @param slave The slave process that gets the System.loadLibrary event.
   */
  private synchronized void onSlaveLoadLibrary(SlaveProcessHandler slave) {
    if (slave != jdbHandler) {
      return;
    }

    new Thread("loadLibraryHandler") {
      public void run() {
        try {
          changeDebugStatus(DebugerControlStatus.JDB);
          runAndWaitJDB("clear java.lang.System.loadLibrary\n",
              "Removed: breakpoint java.lang.System.loadLibrary");
          runAndWaitJDB("step up\n", "Step completed:");
          if (!IsDebuggerSwitchingInitialized()) {
            preparingDebuggerSwitching();
          }

          if (debuggerCommand.HasDeferredBreakPoint()) {
            debuggerCommand.HandleDeferredBreakPoint();
          }

          runAndWaitJDB("stop in java.lang.System.loadLibrary\n",
              "Set breakpoint java.lang.System.loadLibrary");
          jdbHandler.sendMessage("cont\n");
          changeDebugStatus(DebugerControlStatus.NONE);
        } catch (IOException e) {

        }
      }
    }.start();
  }

  /**
   * Release any system resource create for the debugging. For instance, we want
   * to ensure the jdb and gdb are terminated at the end of this method.
   */
  private void ensureResourceRelease() {
    if (jdbHandler != null) {
      jdbHandler.finish();
    }
    if (gdbHandler != null) {
      gdbHandler.finish();
    }
  }

  /**
   * The jdb break point hit event pattern.
   */
  private final static Pattern jdbBreakPointHitPattern = Pattern
      .compile("Breakpoint hit: \\\"thread=([^\"]+)\\\", ([^,]+), line=([0-9,]+) bci=([0-9]+)");

  /**
   * The gdb break point hit event pattern.
   */
  private final static Pattern gdbBreakPointHitPattern = Pattern
      .compile("Breakpoint [0-9]+, ([^ ]+)");

  /**
   * A slave process handler specifically for stream I/O and the process ending.
   * This handler is an interface between the Jeannie debugger and its slave
   * processes. This handler creates three thread to monitor the process state,
   * the process input stream and the process output stream. Mostly all these
   * threads will be blocked, not consuming CPU cycle too much.
   */
  private class SlaveProcessHandler {

    /**
     * The name of this handler.
     */
    private final String name;

    /**
     * The command line to create the native process.
     */
    private final String command;

    /**
     * The native process.
     */
    private Process process;

    /**
     * The buffered writer to the process.
     */
    private BufferedWriter out;

    /**
     * A thread to redirect the process output message to the Jeannie debugger.
     */
    private Thread outRedirector;

    /**
     * A thread to redirect the process error message to the Jeannie debugger.
     */
    private Thread errRedirector;

    /**
     * If set, the output from the slave will be ignored.
     */
    private boolean skipBytesInRedirection;

    /**
     * The expected message from the output stream.
     */
    private Pattern[] expectedMessagePatterns;

    /**
     * The result of expected message matching;
     */
    private DebuggerResponse debuggerResponse;

    /**
     * This buffer keeps a output line whose length is less then 1024. The
     * process handler scans the output from the target process, and try to
     * recognize a line with less than 1024 bytes. If the line over flows, then
     * this handler will ignore the line.
     */
    private final byte[] lineBuffer = new byte[1024];

    /**
     * This state keeps track of the current line output. 0, ...,
     * (jdbLineOutputBuffer.length-1): reading the current line.
     * jdbLineOutputBuffer.length, ...: ignoring the current line.
     */
    private int numBytesInLineBuffer;

    /**
     * The current process state.
     */
    private ProcessState state = ProcessState.NOT_CREATED;

    /**
     * @param jeannieDebugger The Jeannie master debugger.
     * @param name The name of the slave process.
     * @param command The command line to create the slave process.
     */
    SlaveProcessHandler(String name, String command) {
      this.name = name;
      this.command = command;
    }

    /**
     * Start the monitoring thread.
     */
    void start() {
      assert state == ProcessState.NOT_CREATED;
      try {
        process = Runtime.getRuntime().exec(command);
        state = ProcessState.LIVE;
        OutputStream os = process.getOutputStream();
        OutputStreamWriter ow = new OutputStreamWriter(os);
        out = new BufferedWriter(ow);
        outRedirector = new Thread(name + "-out") {
          public void run() {
            handleOutputStream();
          }
        };
        errRedirector = new Thread(name + "-err") {
          public void run() {
            handleErrorStream();
          }
        };
        outRedirector.start();
        errRedirector.start();
      } catch (IOException e) {
        err("failed in executing initj for gdb\n");
        e.printStackTrace();
      }

      assert process != null & outRedirector != null & errRedirector != null;
    }

    /**
     * Complete this slave process.
     */
    void finish() {
      switch (state) {
      case NOT_CREATED:
        state = ProcessState.DEAD;
        break;
      case LIVE:
        state = ProcessState.DEAD;
        process.destroy();
        break;
      case DEAD:
        break;
      default:
        assert false;
        break;
      }

    }

    /**
     * The background routine to redirect the slave output to the Jeannie
     * debugger.
     */
    private void handleOutputStream() {
      InputStream is = process.getInputStream();
      final byte[] buf = new byte[4096];
      try {
        while (true) {
          if (state == ProcessState.DEAD)
            break;
          int nread = is.read(buf);
          if (nread == -1)
            break;
          if (state == ProcessState.DEAD)
            break;

          boolean OKToRedirect;
          synchronized (this) {
            OKToRedirect = !skipBytesInRedirection;
          }
          if (OKToRedirect)
            onSlaveProcessOut(this, buf, nread);
          checkOutputLine(buf, nread);
        }
      } catch (IOException e) {
        if (state == ProcessState.LIVE) {
          System.err.println("IO error in piping thread" + name);
          e.printStackTrace();
        }
      } finally {
        synchronized (this) {
          if (state == ProcessState.LIVE) {
            state = ProcessState.DEAD;
            onSlaveProcessDeath(this);
          }
        }
      }
    }

    /**
     * The background routine to redirect the slave error stream output to the
     * Jeannie debugger.
     */
    private void handleErrorStream() {
      InputStream is = process.getErrorStream();
      final byte[] buf = new byte[4096];
      try {
        while (true) {
          if (state == ProcessState.DEAD)
            break;
          int nread = is.read(buf);
          if (nread == -1) {
            synchronized (this) {
              state = ProcessState.DEAD;
              onSlaveProcessDeath(this);
            }
            break;
          }
          if (state == ProcessState.DEAD)
            break;
          onSlaveProceeErr(this, buf, nread);
        }
      } catch (IOException e) {
        if (state == ProcessState.LIVE) {
          System.err.println("IO error in piping thread" + name);
          e.printStackTrace();
        }
      } finally {
        synchronized (this) {
          if (state == ProcessState.LIVE) {
            state = ProcessState.DEAD;
            onSlaveProcessDeath(this);
          }
        }
      }
    }

    /**
     * Send a message to the output stream.
     * 
     * @param msg The message.
     */
    void sendMessage(String msg) throws IOException {
      out.write(msg);
      out.flush();
    }

    /**
     * Check the read data for the output line event.
     * 
     * @param buf The data buffer.
     * @param len The buffer length.
     */
    void checkOutputLine(byte[] buf, int len) {
      for (int i = 0; i < len; i++) {
        byte c = buf[i];
        if (numBytesInLineBuffer < lineBuffer.length) {
          if (((char) c) == '\n') {
            handleOutputLine(new String(lineBuffer, 0,
                numBytesInLineBuffer));
            numBytesInLineBuffer = 0;
          } else {
            lineBuffer[numBytesInLineBuffer++] = c;
          }
        } else {
          if (((char) c) == '\n') {
            numBytesInLineBuffer = 0;
          } else {
            numBytesInLineBuffer++;
          }
        }
      }
    }

    /**
     * Handle a line from process output.
     * 
     * @param line The line output from the process.
     */
    void handleOutputLine(String line) {
      if (line.equals(""))
        return;

      // handle any expected message request at first.
      checkExpect(line);

      Matcher mjdbbp = jdbBreakPointHitPattern.matcher(line);
      if (mjdbbp.find()) {
        String method = mjdbbp.group(2);
        if (method.startsWith("java.lang.System.loadLibrary")) {
          onSlaveLoadLibrary(this);
        } else if (method.startsWith("DebugHelper")) {

        } else {
          onSlaveProcessBreakPointHit(this);
        }
      }

      Matcher mgdbbp = gdbBreakPointHitPattern.matcher(line);
      if (mgdbbp.find()) {
        onSlaveProcessBreakPointHit(this);
      }
    }

    /**
     * Send a message to the slave process, and wait until the slave sends back
     * an expected message.
     * 
     * @param msg The message to send.
     * @param expect The message to be expected from the slave process.
     */
    Matcher runAndWait(String msg, String expect) throws IOException {
      requestExpect(expect);
      sendMessage(msg);
      waitForExpect();
      return getExpectedMatcher();
    }

    /**
     * Send a message to the slave process, and wait until the slave sends back
     * an expected message.
     * 
     * @param msg The message to send.
     * @param expects The message to be expected from the slave process.
     */
    DebuggerResponse runAndWait(String msg, String[] expects)
        throws IOException {
      requestExpect(expects);
      sendMessage(msg);
      waitForExpect();
      return getDebuggerResponse();
    }

    /**
     * Send a message to this slave process, and wait until the other slave
     * process respond with the expected message.
     * 
     * @param msg The message to send.
     * @param expect The expected message.
     * @param other The other slave process.
     */
    Matcher runAndWait(String msg, String expect, SlaveProcessHandler responder)
        throws IOException {
      responder.requestExpect(expect);
      sendMessage(msg);
      responder.waitForExpect();
      return responder.getExpectedMatcher();
    }

    /**
     * Check if the received bytes says that the expected message was received.
     * 
     * @param line The line output from the output stream.
     */
    private synchronized void checkExpect(String line) {
      if (expectedMessagePatterns == null)
        return;

      for (int i = 0; i < expectedMessagePatterns.length; i++) {
        Pattern p = expectedMessagePatterns[i];
        Matcher m = p.matcher(line);
        if (m.find()) {
          this.expectedMessagePatterns = null;
          this.debuggerResponse = new DebuggerResponse(i, m);
          this.notify();
          return;
        }
      }
    }

    /**
     * Request a message arrival notification.
     * 
     * @param rexpr The regular expression to wait for.
     */
    synchronized void requestExpect(String rexpr) {
      requestExpect(new String[] { rexpr });
    }

    /**
     * 
     * @param rexprs The array of regular repressions to wait for.
     */
    synchronized void requestExpect(String[] exprs) {
      int numRegularExpressions = exprs.length;
      expectedMessagePatterns = new Pattern[numRegularExpressions];
      for (int i = 0; i < numRegularExpressions; i++) {
        String rexpr = exprs[i];
        expectedMessagePatterns[i] = Pattern.compile(rexpr);
      }
      skipBytesInRedirection = true;
    }

    /**
     * Wait until we see the expected message in the stream.
     */
    synchronized void waitForExpect() {
      if (expectedMessagePatterns == null)
        return;
      try {
        this.wait(); // wait until awaken by the checkExpect.
      } catch (InterruptedException e) {
      }
      skipBytesInRedirection = false;
    }

    /**
     * Return the match object for the expected message.
     * 
     * @return The match object for the regular expression
     */
    synchronized Matcher getExpectedMatcher() {
      assert debuggerResponse != null;
      return debuggerResponse.getMatcher();
    }

    /**
     * Return the debugger's response object.
     * 
     * @return The debugger's response.
     */
    synchronized DebuggerResponse getDebuggerResponse() {
      return debuggerResponse;
    }
  }

  /**
   * A debugger's response to the blink's command.
   *  
   */
  public static class DebuggerResponse {

    /**
     * The message ID to show which expected message request. 
     */
    final int expectedMessageID;

    /**
     * The matched message.
     */
    final Matcher matchedMessage;

    /**
     * @param mid The message id.
     * @param msg THe matcher.
     */
    DebuggerResponse(int mid, Matcher msg) {
      expectedMessageID = mid;
      matchedMessage = msg;
    }

    /**
     * @return The expected message ID.
     */
    int getExpectedMessageID() {
      return expectedMessageID;
    }

    /**
     * @return The matcher object.
     */
    Matcher getMatcher() {
      return matchedMessage;
    }
  }
}
