package xtc.lang.jeannie;

import java.io.IOException;
import java.util.regex.Matcher;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.TreeSet;

import xtc.lang.jeannie.Debugger.DebugerControlStatus;
import xtc.tree.GNode;

/**
 * The Blink macro command implementation.
 * 
 * @author Byeongcheol Lee
 */
public final class DebuggerCommand {

  /**
   * The debugger. 
   */
  private final Debugger dbg;

  /**
   * The Jeannie expression handler. 
   */
  final DebuggerExpression dbgExpr;

  /**
   * The blink break point list.
   */
  private final HashMap<Integer,BlinkBreakPoint> breakPoints = 
    new HashMap<Integer,BlinkBreakPoint>();

  /**
   * The deferred break point list.
   */
  private final HashMap<Integer,BlinkNativeBreakPoint> deferredBreakPoints =
    new HashMap<Integer,BlinkNativeBreakPoint>();

  /**
   * @param dbg The debugger to work with.
   */
  DebuggerCommand(Debugger dbg) {
    this.dbg = dbg;
    this.dbgExpr = new DebuggerExpression(dbg);
  }


  /**
   * Extract the current source file and line by asking the current debugger
   * (jdb or gdb).
   * 
   * @return An object containing the source file and the line.
   */
  DebuggerSymbolMapper.SourceFileAndLine getCurrentSourceFileAndLine() {
    // This would be pretty naive implementation.
    switch (dbg.getControlStatus()) {
    case JDB:
      try {
        Matcher m = dbg.runAndWaitJDB("where\n",
            "\\[\\d+\\] [^ ]+ \\(([^:]+):(\\d+)\\)");
        String file = m.group(1);
        int line = Integer.parseInt(m.group(2));
        return new DebuggerSymbolMapper.SourceFileAndLine(file, line);
      } catch (IOException e) {
        dbg.err("can not handle ensureCurrentSourceFileAndLine");
      }
      break;
    case GDB:
      try {
        Matcher m = dbg.runAndWaitGDB("where 1\n", "at ([^:]+):(\\d+)");
        String file = m.group(1);
        int line = Integer.parseInt(m.group(2));
        return new DebuggerSymbolMapper.SourceFileAndLine(file, line);
      } catch (IOException e) {
        dbg.err("can not handle ensureCurrentSourceFileAndLine");
      }
      break;
    default:
      dbg.err("can not handle ensureCurrentSourceFileAndLine");
      break;
    }
    return null;
  }

  /**
   * Pass a command to the jdb.
   * 
   * @param jdbCommand The jdb command.
   */
  void executeJDBCommand(String jdbCommand) {
    if (dbg.getControlStatus() != Debugger.DebugerControlStatus.JDB
        && dbg.getControlStatus() != Debugger.DebugerControlStatus.JDB_IN_GDB) {
      dbg.err("jdb prefix command is allowed only in the jdb mode");
    }
    try {
      dbg.sendUserMessage(jdbCommand + "\n");
    } catch (IOException e) {
      dbg.err("can not successfully jdb command");
    }
  }

  /**
   * Pass a command to the gdb.
   * 
   * @param gdbCommand THe gdb command.
   */
  void executeGDBCommand(String gdbCommand ) {
  if (dbg.getControlStatus() != Debugger.DebugerControlStatus.GDB
        && dbg.getControlStatus() != Debugger.DebugerControlStatus.GDB_IN_JDB) {
      dbg.err("gdb prefix command is allowed only in the gdb mode");
    }
    try {
      dbg.sendUserMessage(gdbCommand + "\n");
    } catch (IOException e) {
      dbg.err("can not successfully gdb command");
    }       
  }



  /**
   * Set a C break point.
   * 
   * @param sourceFile The source file name.
   * @param lineNumber The line number.
   */
   void executeBreak(String sourceFile, int lineNumber) {
     BlinkNativeBreakPoint bp = addNativeBreakPoint(sourceFile, lineNumber);
     if (!dbg.IsDebuggerSwitchingInitialized()) {
       deferredBreakPoints.put(bp.getID(), bp);
       dbg.out("the break point is delayed until the shared library loading\n");
       return;
     }
     bp.set(dbg);
     
  }

  /**
   * Set a Java break point.
   * 
   * @param classID The class name.
   * @param lineNumber The line number.
   */
   void executeStopAt(String classID, int lineNumber) {
    BlinkJavaBreakPoint bp = addJavaBreakPoint(classID, lineNumber);
    bp.set(dbg);
  }

  /**
   * Implement "info break" command.
   */
   void executeInfoBreak() {
     TreeSet<Integer> breakPointIDList = new TreeSet<Integer>();
     breakPointIDList.addAll(breakPoints.keySet());
     StringBuilder sb = new StringBuilder();
     for (final Integer id : breakPointIDList) {
       BlinkBreakPoint bp = breakPoints.get(id);
       sb.append(bp.toString()).append('\n');
     }
     dbg.out(sb.toString());
  }

  /**
   * Implement "delete [n]" command.
   * @param id The identifier of break point or watch point.
   */
   void executeDelete(int id) {
     if (deferredBreakPoints.containsKey(new Integer(id))) {
       deferredBreakPoints.remove(new Integer(id));
       breakPoints.remove(new Integer(id));
       return;
     } else if (!breakPoints.containsKey(new Integer(id))) {
       dbg.err("not valid break point id -" + id);
       return;
     }

     BlinkBreakPoint bp = breakPoints.get(new Integer(id));
     bp.reset(dbg);
     breakPoints.remove(new Integer(id));
   }

  /**
   * Implement the access watch command.
   * 
   * @param accessType The access type. Either "access" or "all"
   * @param className The class name to watch.
   * @param fieldName The field name to watch.
   */
   void executeWatch(String accessType, String className, String fieldName) {
     assert false : "not yet implemented";
  }

  /**
   * Implement the expression watch command. 
   * @param expr The expression
   */
   void executeWatch(String expr) {
     assert false : "not yet implemented";
  }

  /**
   * Implement "info watch" command.
   */
   void executeInfoWatch() {
     assert false : "not yet implemented";
  }

  /**
   * Implement Blink "where" command. 
   */
   void executeWhere() {
    switch(dbg.getControlStatus()) {
    case NONE:
      dbg.out("not available at this time.");
      break;
    case JDB:
    case JDB_IN_GDB:
      try {
        dbg.sendUserMessage("where\n");
      } catch (IOException e) {
        dbg.err("could not correctly run where");
      }
      break;
    case GDB:
    case GDB_IN_JDB:
      try {
        dbg.sendUserMessage("where\n");
      }catch (IOException e) {
        dbg.err("could not correctly run where");
      }
      break;
    default:
      break;
    }
  }

  /**
   * Implement "up [n]" command.
   * @param step The number of stack walk-up.
   */
   void executeUp(int step) {
     assert false : "not yet implemented";
  }

  /**
   * Implement "down [n]" command.
   * 
   * @param step The number of stack walk-down.
   */
   void executeDown(int step) {
     assert false : "not yet implemented";
  }

  /**
   * Implement "locals" command.
   */
   void executeLocals() {
     assert false : "not yet implemented";
  }


  /**
   * Implement the step command.
   */
  void executeStep() {
    assert false : "not yet implemented";
  }

  /**
   * Implement the next command.
   */
   void executeNext() {
    assert false : "not yet implemented";
  }

  /**
   * Implement the print <expr> command.
   * 
   * @param expr The expression.
   */
  void executePrint(final GNode n) {
    DebuggerSymbolMapper.SourceFileAndLine loc = getCurrentSourceFileAndLine();
    switch(dbg.getControlStatus()) {
    case JDB:
      String jdbExpr = dbgExpr.toJDBExpression(n, loc);
      try {
        Matcher m = dbg.runAndWaitJDB("print " + jdbExpr, "(.+ = )");
        dbg.out(m.group(1));
      } catch(IOException e) {
        dbg.err("can not correctly run print");
      }
      break;
    case GDB:
      String gdbExpr = dbgExpr.toGDBExpression(n, loc);
      try {
        Matcher m = dbg.runAndWaitJDB("print " + gdbExpr, "(.+ = )");
        dbg.out(m.group(1));
      }catch(IOException e) {
        dbg.err("can not correctly run print");
      }
      break;
    default:
      assert false: "not suitable debugger status";
      break;
    }
    
  }

  /**
   * Implement Blink "list" command.
   */
  void executeList() {
    try {
      dbg.sendUserMessage("list\n");
    } catch (IOException e) {
      dbg.err("could not run the list ");
    }
  }

  /**
   * @return true if there is blink deferred C break point.
   */
  boolean HasDeferredBreakPoint() {
    return deferredBreakPoints.size() > 0;
  }

  /**
   * Handle deferred C break points when the shared library is loaded into the
   * debuggee JVM.
   */
  void HandleDeferredBreakPoint() {
    assert dbg.getControlStatus() == Debugger.DebugerControlStatus.GDB
        || dbg.getControlStatus() == Debugger.DebugerControlStatus.JDB;
    for (BlinkNativeBreakPoint p : deferredBreakPoints.values()) {
      p.set(dbg);
    }
    deferredBreakPoints.clear();
  }

  /**
   * 
   * @param sourceFile The source file name.
   * @param sourceLine The source line number.
   */
  private BlinkNativeBreakPoint addNativeBreakPoint(String sourceFile, int sourceLine) {
    BlinkNativeBreakPoint nbp =new BlinkNativeBreakPoint(sourceFile, sourceLine);
    breakPoints.put(new Integer(nbp.getID()), nbp);
    return nbp;
  }

  /**
   * 
   * @param classFile The class name.
   * @param sourceLine The source line number.
   */
  private BlinkJavaBreakPoint addJavaBreakPoint(String classFile, int sourceLine) {
    BlinkJavaBreakPoint jbp = new BlinkJavaBreakPoint(classFile, sourceLine);
    breakPoints.put(new Integer(jbp.getID()), jbp);
    return jbp;
  }

   /**
     * A base class for the Blink break point.
     */
   private static abstract class BlinkBreakPoint {
     
     /** 
      * The next identification sequence number.
      */
     private static int nextID = 1;
     
     /**
      * Allocate a new identification number.
      * 
      * @return A new identification number.
      */
     private static synchronized int getNextID() {
       return nextID++;
     }

     /**
      * The identification number.
      */
     private final int Id;
     
     protected BlinkBreakPoint() {
       this.Id = getNextID();
     }

     /**
      * @return The identification number of the break point.
      */
     public int getID() {
       return Id;
     }
     
     abstract void set(Debugger dbg);
     abstract void reset(Debugger dbg);
   }

   /**
   * A deferred break point in the native code.
   */
   private static class BlinkNativeBreakPoint extends BlinkBreakPoint {

    /**
     * The source file name.
     */
     final String sourceFileName;

    /**
     * The line number in the source file.
     */
     final int sourceLineNumber;

    /**
     * @param sname The source file name.
     * @param lineno The line number.
     */
     BlinkNativeBreakPoint(String sname, int lineno) {
      sourceFileName = sname;
      sourceLineNumber = lineno;
    }

    /**
     * @return The source file name.
     */
    String getSourceFileName() {
      return sourceFileName;
    }

    /**
     * @return The source line number.
     */
    int getSourceLineNumber() {
      return sourceLineNumber;
    }
    

    /**
     * Check the equality of the native break point.
     *  
     * @param o The compared object.
     */
    public boolean equals(Object o) {
      if (o instanceof BlinkNativeBreakPoint == false) return false;
      BlinkNativeBreakPoint nbp = (BlinkNativeBreakPoint)o;
      return (this == nbp)
          || (this.sourceFileName.equals(nbp.sourceFileName) && this.sourceLineNumber == nbp.sourceLineNumber);
    }

    /**
     * @return The string representation.
     */
    public String toString() {
      StringBuilder sb = new StringBuilder();
      sb.append(getID()).append(" ").append("native").append(" ");
      sb.append(sourceFileName).append(":").append(sourceLineNumber);
      return sb.toString();
    }
    
    void set(Debugger dbg) {

      switch(dbg.getControlStatus()) {
      case JDB:
          try {
          dbg.j2n();
          // try to set break point
          dbg.runAndWaitGDB("break " + sourceFileName + ":" + sourceLineNumber
              + "\n", "Breakpoint [0-9]+ at");
          dbg.jret();
        } catch (IOException e) {
          dbg.err("could not set the break point.");
        }
        break;
      case GDB:
        try {
          //try to set the break point
          //try to set break point
          dbg.runAndWaitGDB(
              "break " + sourceFileName + ":" + sourceLineNumber + "\n", 
              "Breakpoint [0-9]+ at");
        }catch(IOException e) {
          dbg.err("could not set the break point.");
        }        
        break;
      default:
        break;
      }
    }
    
    /**
     * Delete this break in the gdb.
     * @param dbg The Blink debugger.
     */
    void reset(Debugger dbg) {
      assert (dbg.getControlStatus() == Debugger.DebugerControlStatus.JDB)
          || (dbg.getControlStatus() == Debugger.DebugerControlStatus.GDB);
      assert (dbg.IsDebuggerSwitchingInitialized());
      boolean needSwitching = dbg.getControlStatus() == Debugger.DebugerControlStatus.JDB;
      if (needSwitching) dbg.j2n();
      
      try {
        dbg.runAndWaitGDB(
            "clear " + sourceFileName + ":"  + sourceLineNumber + "\n",
            "Deleted breakpoint [0-9]+");
      } catch(IOException e) {
        dbg.err("cound not resent the break point");
      }
      
      if (needSwitching) dbg.jret();
    }
  }
   
   /**
    * A deferred break point in the native code.
    */
    private static class BlinkJavaBreakPoint extends BlinkBreakPoint {

     /**
      * The source file name.
      */
      final String className;

     /**
      * The line number in the source file.
      */
      final int sourceLineNumber;

     /**
      * @param sname The source file name.
      * @param lineno The line number.
      */
      BlinkJavaBreakPoint(String sname, int lineno) {
       className = sname;
       sourceLineNumber = lineno;
     }

     /**
      * @return The source file name.
      */
     String getClassName() {
       return className;
     }

     /**
      * @return The source line number.
      */
     int getSourceLineNumber() {
       return sourceLineNumber;
     }
     /**
      * Check the equality of the native break point.
      *  
      * @param o The compared object.
      */
     public boolean equals(Object o) {
       if (o instanceof BlinkNativeBreakPoint == false) return false;
       BlinkJavaBreakPoint nbp = (BlinkJavaBreakPoint)o;
       return (this == nbp)
           || (this.className.equals(nbp.className) && this.sourceLineNumber == nbp.sourceLineNumber);
     }

     /**
      * @return The string representation.
      */
     public String toString() {
       StringBuilder sb = new StringBuilder();
       sb.append(getID()).append(" ").append("java").append(" ");
       sb.append(className).append(":").append(sourceLineNumber);
       return sb.toString();
     }
     
     void set(Debugger dbg) {
       switch(dbg.getControlStatus()) {
       case JDB:
         try {
           //try to set the break point
           dbg.runAndWaitJDB(
               "stop at " + className + ":" + sourceLineNumber + "\n", 
               new String[] {
                 "Deferring breakpoint [^ ]+:[0-9]+.",
                 "Set breakpoint [^ ]+:[0-9]+",
               });
         }catch(IOException e) {
           dbg.err("could not set the break point.");
         }              
         break;
       case GDB:
         try {
           dbg.n2j();
           dbg.runAndWaitJDB(
               "stop at " + className + ":" + sourceLineNumber + "\n", 
               new String[] {
                 "Deferring breakpoint [^ ]+:[0-9]+.",
                 "Set breakpoint [^ ]+:[0-9]+",
               });
           dbg.jret();
         }catch(IOException e) {
           dbg.err("could not set the break point.");
         }
         break;
       default:
         break;
       }
     }
     /**
      * Delete this break in the gdb.
      * @param dbg The Blink debugger.
      */
     void reset(Debugger dbg) {
       assert (dbg.getControlStatus() == Debugger.DebugerControlStatus.JDB)
          || (dbg.getControlStatus() == Debugger.DebugerControlStatus.GDB);
      assert (dbg.IsDebuggerSwitchingInitialized());
      boolean needSwitching = dbg.getControlStatus() == Debugger.DebugerControlStatus.GDB;
      if (needSwitching)
        dbg.n2j();

      try {
        dbg.runAndWaitJDB("clear " + className + ":" + sourceLineNumber
            + "\n", "Removed: breakpoint ([^ ]+):([0-9]+)");
      } catch (IOException e) {
        dbg.err("cound not resent the break point");
      }

      if (needSwitching)
        dbg.jret();       
     }
   }
}
