package xtc.lang.jeannie;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import xtc.tree.GNode;

/**
 * A Jeannie expression handler for the Blink debugger.
 *  
 * @author Byeongcheol Lee
 */
public class DebuggerExpression implements DebuggerSymbolMapper {
  enum TargetSourceLanguage {C, JAVA};

  /**
   * The debugger to contact to evaluate Jeannie expression.
   */
  private final Debugger dbg;

  /**
   * The Jeannie source level variable remapping. The blink debugger search the
   * source file direction for the variable remapping table. For instance, the
   * blink debugger expect to find a remap file, "Main.jni.symbol" for the
   * source file "Main.jni."
   */
  private final HashMap<String, DebuggerExpression.SourceVariableMapper> varibleRemap
  = new HashMap<String, DebuggerExpression.SourceVariableMapper>();

  /**
   * @param dbg The debugger.
   */
  DebuggerExpression(Debugger dbg) {
   this.dbg = dbg;
  }


  /**
   * Tries to find the Jeannie variable remap information.
   * 
   * @param variable The variable name.
   * @param sourceFile The source file.
   * @param sourceLineNumber The source file line number.
   * @return a variable remapper object if sucessful otherwise null.
   */
  public VariableRemapEntry lookUpVariableRemap(String variable,
      String sourceFile, int sourceLineNumber) {
    if (sourceFile == null || sourceLineNumber < 0)
      return null;

    DebuggerExpression.SourceVariableMapper mapper = null;
    if (!varibleRemap.containsKey(sourceFile)) {
      //the first time
      try {
        String remapFile = sourceFile + ".symbols";
        mapper = new DebuggerExpression.SourceVariableMapper(remapFile);
      } catch (IOException e) {
      }
    } else {
      mapper = varibleRemap.get(sourceFile);
    }
    if (mapper == null)
      return null;

    return mapper.lookUp(variable, sourceLineNumber);
  }

  /**
   * A source level variable remapping table for a source file.
   */
  private static class SourceVariableMapper {

    /**
     * The remap file.
     */
    private final String remapFile;

    /**
     * The list of variable remap entry.
     */
    private final List<VariableRemapEntry> entries = new LinkedList<VariableRemapEntry>();

    /**
     * @param remapFile The file containing the variable remapping information.
     */
    public SourceVariableMapper(String remapFile) throws IOException {
      this.remapFile = remapFile;
      updateRemapEntry();
    }

    /**
     * Update the remap entry from the remap file.
     */
    private void updateRemapEntry() throws IOException {
      entries.clear();

      Pattern remapEntryPattern = Pattern
          .compile("\\s*(\\S+) (\\d+) (\\d+) (\\d+) (\\d+) (Java|C) (\\S+) (\\S+)");
      BufferedReader br = new BufferedReader(new FileReader(remapFile));
      for(String line = br.readLine(); line != null; line = br.readLine()) {
        Matcher m =remapEntryPattern.matcher(line);
        if (!m.matches()) {continue;}
        int startLine = Integer.parseInt(m.group(2));
        int startColumn = Integer.parseInt(m.group(3));
        int endLine = Integer.parseInt(m.group(4));
        int endColumn = Integer.parseInt(m.group(5));
        String targetLanguageStr = m.group(6);
        TargetSourceLanguage tlang;
        if (targetLanguageStr.equals("Java")) {
          tlang = TargetSourceLanguage.JAVA;
        } else {
          assert targetLanguageStr.equals("C");
          tlang = TargetSourceLanguage.C;
        }
        String sourceVariableName = m.group(7);
        String targetVariableName = m.group(8);
        VariableRemapEntry e = new VariableRemapEntry(startLine, startColumn,
            endLine, endColumn, tlang, sourceVariableName, targetVariableName);
        entries.add(e);
      }
      br.close();
    }

    /**
     * Look up for the variable remap entry.
     * 
     * @param var The variable name.
     * @param sourceLine The source line number for the variable.
     * @return A VariableRemapEntry object if found, or null otherwise.
     */
    VariableRemapEntry lookUp(String var, int sourceLine) {
      for(final VariableRemapEntry e : entries) {
        if (sourceLine >= e.startLine && sourceLine <= e.endLine
            && e.sourceVariableName.equals(var)) {
          return e;
        }
      }
      return null;
    }
  }

  String toJDBExpression(final GNode n, SourceFileAndLine loc) {
    return "";
  }

  String toGDBExpression(final GNode n, SourceFileAndLine loc) {
    return "";
  }

  public SourceFileAndLine getCurrentSourceFileAndLine() {
    return dbg.debuggerCommand.getCurrentSourceFileAndLine();
  }
}
