/*
    Katja
    Copyright (C) 2003-2007 see README file for authors

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package katja.common;

import java.io.IOException;
import java.io.StringReader;

public interface KatjaElement {

    /**
     * Returns the name of the sort of this element.
     * @return the name of the sort of this element
     */
    public String sortName();

    /**
     * Returns the number of children this element has.
     * @return the number of children this element has
     */
    public int size();

    /**
     * Returns a child element specified by its position, starting at 0.
     * Allows backward adressing, i.e. -1 is the last element, up to -size
     * for the first element. A wrong index results in an illegal argument
     * exception.
     * @param i the child element to get
     * @return a child element
     */
    public Object get(int i);

    /**
     * Creates a pretty string represenation of the element (still in one line
     * only, not using newlines or indentation).
     * @return the pretty string representation of this element
     */
    public String toString();

    /**
     * Creates a string representation of the element, which can be used in
     * Java code to recreate the element.
     * @return the java code representation of this element
     */
    public String toJavaCode();

    /**
     * Creates an assembler listing, which can be interepreted by Katja
     * specification classes to recreate the element.
     * @return the assembler representation of this element
     */
    public String toAssembler();

    /**
     * Creates a pretty string represenation of the element (still in one line
     * only, not using newlines or indentation).
     * This version allows to use specified Appendable to append the output to.
     * @return the pretty string representation of this element
     */
    public Appendable toString(Appendable builder) throws IOException;

    /**
     * Creates a string representation of the element, which can be used in
     * Java code to recreate the element.
     * This version allows to use specified Appendable to append the output to.
     * @return the java code representation of this element
     */
    public Appendable toJavaCode(Appendable builder) throws IOException;

    /**
     * Creates an assembler listing, which can be interepreted by Katja
     * specification classes to recreate the element.
     * This version allows to use specified Appendable to append the output to.
     * @return the assembler representation of this element
     */
    public Appendable toAssembler(Appendable builder) throws IOException;

    /**
     * Can be used to cast an element down in a method cascade, i.e. without
     * the need to write the type in front of a expression.
     * @return this
     */
    public <T extends KatjaElement> T cast();

    /**
     * Contains static utilities for generic Katja elements, like pretty
     * printing.
     */
    public static class Utilities {

        /**
         * Generic, inefficient method, taking an existing output string of a term and adding newlines as well as
         * indents. The method works on any string structures with pairs of '(' and ')', so toString as well as
         * toJavaCode output. Best used with smaller terms only, as the output does not really become more pretty for
         * large terms and it becomes more and more inefficient.
         * @param term the string represantation of a term
         * @return the reformated version of the term
         */
        public static String prettyPrint(String term) {

            // create a reader for the string
            StringReader reader = new StringReader(term);

            // create an output builder
            StringBuilder builder = new StringBuilder(term.length()*2);

            // the current indent level
            int tabsize = 0;

            // the current character from term
            int character;

            // wether white spaces should be ignored at the moment or not
            boolean eatws = false;

            // wether the next line of real characters needs indentation
            boolean indent = false;

            try {

                // as long as there is input left, read the next character
                while((character = reader.read()) != -1) {

                    switch(character) {

                        // whitespaces
                        case ' ':
                        case '\n':
                        case '\t':

                            // if in eatws, then eat
                            if(eatws) continue;
                            // otherwise print them, but never indent!
                            else {
                                builder.append((char) character);
                                continue;
                            }

                        // the closing bracket need immediate action, as itself is indented
                        case ')':

                            // reduce tab size, newline and then indent the character itself
                            tabsize--;
                            builder.append('\n');
                            for(int i = 0; i < tabsize; i++) builder.append("  ");

                            // if there was a pending indent, cancel it, eat whitespaces after the ')'
                            indent = false;
                            eatws = true;

                            // continue with default case, i.e. print character, etc.

                        // all other text
                        default:

                            // indentation means a newline and many spaces, according to tabsize, this ends eatws
                            if(indent) {
                                eatws = false;
                                indent = false;
                                builder.append('\n');
                                for(int i = 0; i < tabsize; i++) builder.append("  ");
                            }

                            // append the original character
                            builder.append((char) character);

                    }

                    // see if we need to do anything special
                    switch(character) {

                        // a new component block, indent it
                        case '(':

                            // increase tab size, eat comming whitespaces and newline+indent the next text
                            tabsize++;
                            eatws = true;
                            indent = true;
                            break;

                        // next part of a component block
                        case ',':

                            // eat comming whitespaces and put the next text in an indented newline
                            eatws = true;
                            indent = true;
                            break;

                        // all other text
                        default:

                            // no action in other cases
                    }
                }
            } catch (IOException e) {
                throw new IllegalStateException("StringReader should never throw IOExceptions");
            }

            // return the new string
            return builder.toString();
        }
    }
}
