/*
    Katja
    Copyright (C) 2003-2007 see README file for authors

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package katja.common;

import java.io.IOException;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

public abstract class KatjaElementImpl implements KatjaElement {

    /**
     * Overrides the standard toString to use the special string builder version
     * @return the pretty string representation of this element
     */
    public String toString() {
        try {
            return toString(new StringBuilder()).toString();
        } catch (IOException e) {
            throw new IllegalStateException("StringBuilder should not throw an IOException if used as Appendable");
        }
    }

    /**
     * Uses the string builder version to create the result string
     * @return the java code representation of this element
     */
    public String toJavaCode() {
        try {
            return toJavaCode(new StringBuilder()).toString();
        } catch (IOException e) {
            throw new IllegalStateException("StringBuilder should not throw an IOException if used as Appendable");
        }
    }

    /**
     * Uses the string builder version to create the result string
     * @return the assembler representation of this element
     */
    public String toAssembler() {
        try {
            return toAssembler(new StringBuilder()).toString();
        } catch (IOException e) {
            throw new IllegalStateException("StringBuilder should not throw an IOException if used as Appendable");
        }
    }

    /**
     * Unparse given katja element to the katja assembly language
     * @param builder the appendable to use for the output
     * @return the assembler code producing given term
     * @throws IOException appendable operations may throw an exception
     */
    public Appendable toAssembler(Appendable builder) throws IOException {
        return toAssembler(this, builder, true);
    }

    /**
     * Unparse given katja element to the katja assembly language
     * @param element the element to unparse
     * @param builder the appendable to use for the output
     * @param close wether to close the assembly or not
     * @return the assembler code producing given term
     * @throws IOException appendable operations may throw an exception
     */
    public static Appendable toAssembler(KatjaElement element, Appendable builder, boolean close) throws IOException {

        // if this element is a position, just unparse the term and wrap a constructor
        if(element instanceof KatjaSortPos) {

            toAssemblerInternal(builder, ((KatjaSortPos<?>) element).term(), new ArrayList<KatjaSortPos<?>>());
            builder.append("#apply "+element.sortName()+" 1");
            builder.append('\n');

        // if it's a term there can be loads of different positions from different structure, so we have to unparse
        // all those root terms first
        } else if(element instanceof KatjaSort) {

            // create a root term list
            List<KatjaSortPos<?>> rootStack = new ArrayList<KatjaSortPos<?>>();

            // go through all terms, to search for positions
            Iterator<Object> iterator = ((KatjaSort) element).iterate();
            while(iterator.hasNext()) {
                Object term = iterator.next();

                // if we have found a position
                if(term instanceof KatjaSortPos<?>) {

                    // get the position
                    KatjaSortPos<?> root = ((KatjaSortPos<?>) term).root();

                    // if we don't know the root so far
                    if(!rootStack.contains(root)) {

                        // add the root to the known roots
                        rootStack.add(root);

                        // unparse it once
                        toAssembler(root, builder, false);
                    }
                }
            }

            // now unparse the term
            toAssemblerInternal(builder, element, rootStack);

        // if it's neither a term, nor a position... we are confused
        } else throw new IllegalStateException("found a katja element, which is neither sort nor sort position");

        // mark the end of the program
        if(close) {
            builder.append("#end");
            builder.append('\n');
        }

        // return the appendable for convenience
        return builder;
    }

    /**
     * Unparse given term to assembly code. This method does not need to care for positions, as it
     * is called with a stack of roots, so it can generate load references.
     * @param builder the builder to append the output to
     * @param element the term to unparse
     * @param rootStack the root positions to use
     * @throws java.io.IOException the appendable can throw this
     */
    private static void toAssemblerInternal(Appendable builder, Object element, List<KatjaSortPos<?>> rootStack) throws IOException {

        // if we shall unparse a position
        if(element instanceof KatjaSortPos<?>) {

            // get the position reference
            KatjaSortPos<?> pos = (KatjaSortPos<?>) element;

            // get the index of root on the stack
            int index = rootStack.indexOf(pos.root());

            // if there is no such index... something went wrong
            if(index == -1) throw new IllegalStateException("toAssemblyInternal invoked with term containing a position,"+
                    " whose root is not on the root stack");

            // put the load and get instructions to the file
            builder.append("#load "+index);
            builder.append('\n');
            if(!pos.path().isEmpty()) {

                // return the result line
                builder.append("#follow");

                // print numbers space seperated, with space before the first one
                for(int position : pos.path()) builder.append(" "+position);
                builder.append('\n');

            }

        // so we have a term
        } else if(element instanceof KatjaSort) {

            // get the term reference
            KatjaSort term = (KatjaSort) element;

            // construct all children on the stack, in inverted order
            for(int i = term.size() - 1; i >= 0; i--)
                toAssemblerInternal(builder, term.get(i), rootStack);

            // add constructor call (with arity)
            builder.append("#apply "+term.sortName()+" "+term.size());
            builder.append('\n');

        // so we have an external
        } else {
            if(element instanceof Integer) {
                builder.append("#int "+element);
                builder.append('\n');
            } else if(element instanceof Byte) {
                builder.append("#byte "+element);
                builder.append('\n');
            } else if(element instanceof Short) {
                builder.append("#short "+element);
                builder.append('\n');
            } else if(element instanceof Long) {
                builder.append("#long "+element);
                builder.append('\n');
            } else if(element instanceof Float) {
                builder.append("#float "+element);
                builder.append('\n');
            } else if(element instanceof Double) {
                builder.append("#double "+element);
                builder.append('\n');
            } else if(element instanceof Boolean) {
                builder.append("#boolean "+element);
                builder.append('\n');
            } else if(element instanceof String) {
                builder.append("#string \""+ ((String) element).replaceAll("\\\\", "\\\\\\\\").replaceAll("\\n", "\\\\n")+"\"");
                builder.append('\n');
            } else {
                throw new IllegalArgumentException("unsupported external: "+element);
            }
        }
    }

    /**
     * Can be used to cast an element down in a method cascade, i.e. without
     * the need to write the type in front of a expression.
     * @return this
     */
    public <T extends KatjaElement> T cast() {
        return (T) this;
    }
}
