/*
    Katja
    Copyright (C) 2003-2007 see README file for authors

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package katja.common;

import java.util.*;


/**
 * General list term implementation, to be inherited from lists
 */
public abstract class KatjaListImpl<L> extends KatjaTermImpl implements KatjaList<L> {

    // the array that stores the list elements
    protected L[] values = null;

    // holds the set version of this list, either null if not known so far, this if it is a set, or some other reference
    protected KatjaList<L> set = null;

    /**
     * The list constructor, can be used by list implementations. Not public at the moment, as this type is not
     * intended to be used by common users.
     * @param elements the elements of this list
     */
    protected KatjaListImpl(L[] elements) {

        // add all elements to a new arry
        values = createArray(elements.length);
        System.arraycopy(elements, 0, values, 0, elements.length);
    }

    /**
     * This constructor will be called ONLY if an list object is needed whose value attribute is set afterwards!
     */
    protected KatjaListImpl() {
        // value is NULL!
    }

    /**
     * Needed internal method for all those mutable operations in lists
     * @param elements the elements to place in the new list, without copying
     * @param isSet wether the new instance will known to be a set or not known
     * @return a new list object with the given array as internal array
     */
    protected abstract KatjaList<L> createInstance(L[] elements, boolean isSet);

    /**
     * Does the work of creating an L array, which can only be done in the generated classes, as L is fixed there (statically)
     * @param size the size of the array
     * @return a new array of type L
     */
    protected abstract L[] createArray(int size);

    /**
     * Returns a an array containing the elements of the list, which has to be a copy of the internal array of course.
     * @return a an array containing the elements of the list
     */
    public L[] toArray() {
        return values.clone();
    }

    /**
     * Returns an iterator for the lists elements
     * @return an iterator for the lists elements
     */
    public Iterator<L> iterator() {
        return new Iterator<L>() {
            int current = 0;

            public boolean hasNext() {
                return current < values.length;
            }

            public L next() {
                if(current < values.length) return values[current++];
                else throw new NoSuchElementException();
            }

            public void remove() {
                throw new UnsupportedOperationException("remove on KatjaList iterator called");
            }
        };
    }

    /**
     * Returns the ith element of the list, backward indexation with negative indexes allowed
     * @param i the component to get, negative values index from the end of the list
     * @return the requested element of the list
     */
    public L get(int i) {

        // save original parameter value for error message
        int ith = i;

        // if negative we assume indexation from the end of the list
        if(ith < 0) ith += size();

        // if still negative it's out of bounds
        if(ith < 0)
            throw new IndexOutOfBoundsException("get on list sort invoked with negative parameter "+i);

        // if too large it's out of bounds too
        if(ith >= values.length)
            throw new IndexOutOfBoundsException("get on list sort invoked with too large parameter "+i+
                    ", list has only "+values.length+" elements");

        // return the requested value
        return values[ith];
    }

    /**
     * Returns the size of the list.
     * @return the size of the list
     */
    public int size() {
        return values.length;
    }

    /**
     * Returns the first element of the list, null if none
     * @return the first element of the list
     */
    public L first()  {

        // the operation is not defined on empty lists
        if(values.length == 0)
            throw new UnsupportedOperationException("first invoked on empty list");

        // returns the first element
        return values[0];
    }

    /**
     * Returns the last element of the list, null if none
     * @return the last element of the list
     */
    public L last()  {

        // the operation is not defined on empty lists
        if(values.length == 0)
            throw new UnsupportedOperationException("last invoked on empty list");

        // returns the last element
        return values[values.length - 1];
    }

    /**
     * Returns the list without first element, not defined on empty lists
     * @return the list without first element
     */
    public KatjaList<L> front()  {

        // the operation is not defined on empty lists
        if(values.length == 0)
            throw new UnsupportedOperationException("front invoked on empty list");


        // copy all values but the last into the list
        L[] temp = createArray(values.length - 1);
        System.arraycopy(values, 0, temp, 0, values.length - 1);

        // returns the front elements
        return createInstance(temp, this.equals(set));
    }

    /**
     * Returns the list without last element, not defined on empt lists
     * @return the list without last element
     */
    public KatjaList<L> back()  {

        // the operation is not defined on empty lists
        if(values.length == 0)
            throw new UnsupportedOperationException("back invoked on empty list");

        // copy all values but the first into the list
        L[] temp = createArray(values.length - 1);
        System.arraycopy(values, 1, temp, 0, values.length - 1);

        // returns the front elements
        return createInstance(temp, this.equals(set));
     }

    /**
     * Adds an element to the list by appending it at the end of the list
     * @param element the element to add
     * @return a new list with the element added
     */
    public KatjaList<L> add(L element) {
        return appBack(element);
    }

    /**
     * Adds all given elements to the list, by appending them at the end of the list, preserving order
     * @param list the list of elements to add
     * @return a new list with all elements added
     */
    public KatjaList<L> addAll(KatjaList<? extends L> list) {
        return conc(list);
    }

    /**
     * Removes an element from the list by deleting exactly one occurrence, which is the first in the list,
     * undefined operation if the element does not exist.
     * @param element the element to remove
     * @return a new list with a given element removed
     */
    public KatjaList<L> remove(L element)  {

        // if the list has no elements return null
        if(values.length == 0)
            throw new IllegalArgumentException("remove invoked on list that does not contain the element at least once");

        // get the position of the first occurrence of element in this list
        int position = Arrays.asList(values).indexOf(element);

        // if the list does not contain such an element throw an exception
        if(position == -1)
            throw new IllegalArgumentException("remove invoked on list that does not contain the element at least once");

        // create a new list which does not contain this one element
        L[] temp = createArray(values.length - 1);
        System.arraycopy(values, 0, temp, 0, position);
        System.arraycopy(values, position + 1, temp, position, values.length - position - 1);

        // and return a new list missing the element
        return createInstance(temp, this.equals(set));
    }

    /**
     * Removes all given elements from the list by deleting all occurrences of all elements from the lust, returns
     * this if the list did not contain any elements at all.
     * @param list the list of elements to remove
     * @return a new list with all given elements removed
     */
    public KatjaList<L> removeAll(KatjaList<? extends L> list)  {

        // if the list has no elements at all return the list
        if(values.length == 0) return this;

        // if the given list has no elements at all return this
        if(list.size() == 0) return this;

        // delete all elements from this list that appear in the given list
        KatjaList<L> temp = this;

        // go through all elemnts of this list, as often as they appear
        for(L element : list.toSet())
            while(temp.contains(element)) temp = temp.remove(element);

        // return the filtered list
        return temp;
    }

    /**
     * Adds an element at the front of the list
     * @param element the element to add
     * @return a new list with the element added at the front of the list
     */
    public KatjaList<L> appFront(L element)  {

        // create a larger array
        L[] temp = createArray(values.length + 1);

        // copy existing values
        System.arraycopy(values, 0, temp, 1, values.length);

        // add the new one
        temp[0] = element;

        // return a new instance
        return createInstance(temp, this.equals(set) && !this.contains(element));
    }

    /**
     * Adds an element at the back of the list
     * @param element the element to add
     * @return a new list with the element added at the back of the list
     */
    public KatjaList<L> appBack(L element)  {

        // create a larger array
        L[] temp = createArray(values.length + 1);

        // copy existing values
        System.arraycopy(values, 0, temp, 0, values.length);

        // add the new one
        temp[values.length] = element;

        // return a new instance
        return createInstance(temp, this.equals(set) && !this.contains(element));
    }

    /**
     * Concatenates two lists with each other
     * @param list the list to add to this list
     * @return a new list representing the concatenation of both lists
     */
    public KatjaList<L> conc(KatjaList<? extends L> list)  {

        // create a new list of correct size
        L[] temp = createArray(values.length + list.size());

        // add all elements of the first list
        System.arraycopy(values, 0, temp, 0, values.length);

        // add all elements of the second list
        System.arraycopy(list.toArray(), 0, temp, values.length, list.size());

        // create the result list
        return createInstance(temp, false);
    }

    /**
     * Returns wether this list has no elements or not
     * @return wether this list has no elements or not
     */
    public boolean isEmpty() {
        return values.length == 0;
    }

    /**
     * Returns wether this list contains the given element or not
     * @param element the element to check
     * @return wether this list contains the given element or not
     */
    public boolean contains(L element) {
        return Arrays.asList(values).contains(element);
    }

    /**
     * Returns wether this list contains all of the given elements or not
     * @param list the elements to check
     * @return wether this list contains all of the given elements or not
     */
    public boolean containsAll(KatjaList<? extends L> list) {
        for(L element : list)
            if(!this.contains(element)) return false;

        return true;
    }

    /**
     * Returns the reversed list
     * @return the reversed list
     */
    public KatjaList<L> reverse() {

        // create an array of same length
        L[] temp = createArray(values.length);

        // go through this list and add the elements reversed to the other list (this first element is length - 1)
        int i = 1;
        for(L element : values) {
            temp[values.length - i] = element;
            i++;
        }

        // return the new unique list
        return createInstance(temp, this.equals(set));
    }

    /**
     * Returns the list containing only a sublist of the elements of this list
     * @param from the starting index of the sublist, includung this element
     * @param to the ending index of the sublist, exluding this element
     * @return the list containing only a sublist of the elements of this list
     */
    public KatjaList<L> sublist(int from, int to) {

        // preserve parameter values for later
        int sfrom = from;
        int sto = to;

        // allow negative addressing
        if(sfrom < 0) sfrom += values.length;
        if(sto < 0) sto += values.length;

        // check bounds
        if(sfrom < 0 || sfrom >= values.length)
            throw new IndexOutOfBoundsException("sublist on list sort invoked with 'from' parameter out of bounds "+from);
        if(sto < 0 || sto > values.length)
            throw new IndexOutOfBoundsException("sublist on list sort invoked with 'to' parameter out of bounds "+to);

        // create array of correct size
        L[] temp;

        // allow overflow addressing
        if(sfrom <= sto) {
            temp = createArray(sto - sfrom);
            System.arraycopy(values, sfrom, temp, 0, sto - sfrom);
        } else {
            temp = createArray(sto - sfrom + values.length);
            System.arraycopy(values, sfrom, temp, 0, values.length - sfrom);
            System.arraycopy(values, 0, temp, values.length - sfrom, sto);
        }

        // return the new instance
        return createInstance(temp, this.equals(set));
    }

    /**
     * Returns the set version of this list, i.e. duplicate items removed
     * @return the set version of this list
     */
    public KatjaList<L> toSet() {

        // check if this list is already known to be a set, if so: return it
        if(set != null) return set;

        // if the list is empty or has only one element: just return it
        if(values.length < 2) {
            set = this;
            return this;
        }

        // go through the list and add new elements
        List<L> list = new ArrayList<L>(values.length);
        for(L element : values) if(!list.contains(element)) list.add(element);
        set = createInstance(list.toArray(createArray(list.size())), true);

        // return the new list (or even this through term sharing)
        return set;
    }

    /**
     * Returns the set version of this list, with the given element added, if new to the list
     * @param element the element to add
     * @return the set version of this list, with the given element added, if new to the list
     */
    public KatjaList<L> setAdd(L element) {

        // be sure we know how a set of our elements looks like
        if(set == null) toSet();

        // add the element, if not already in
        if(!set.contains(element)) return set.add(element);
        else return set;
    }

    /**
     * Returns the set version of this list, but only those elements only contained in the given list
     * @param list the list to intersect with
     * @return the set version of this list, but only those elements only contained in the given list
     */
    public KatjaList<L> setIntersection(KatjaList<? extends L> list) {

        // if we take all elements in list from this list and remove those elements from this: we have the intersection
        // we remove all elements, which where NOT in list from this list
        return this.setWithout(this.setWithout(list));
    }

    /**
     * Returns the set version of this list, with one element removed, does NOT throw an exception if the
     * element is not contained in the set
     * @param element the elemenet to remove
     * @return the set version of this list, with one element removed
     */
    public KatjaList<L> setRemove(L element) {

        // be sure we know how a set of our elements looks like
        if(set == null) toSet();

        // return the new set
        if(!set.contains(element)) return set;
        else return set.remove(element);
    }

    /**
     * Returns the set version of this list, with all given elements added, if not already in
     * @param list the list of elements to add
     * @return the set version of this list, with all given elements added, if not already in
     */
    public KatjaList<L> setUnion(KatjaList<? extends L> list) {

        // return the new union set
        return this.addAll(list).toSet();
    }

    /**
     * Returns the set version of this list, with all given elements removed, if they are in
     * @param list the list of elements to remove
     * @return the set version of this list, with all given elements removed, if they are in
     */
    public KatjaList<L> setWithout(KatjaList<? extends L> list) {

        // be sure we know how a set of our elements looks like
        if(set == null) toSet();

        // return the new set
        return set.removeAll(list);
    }

    /**
     * Returns wether the set version of this list, contains exactly the elements in the given list, seen as set.
     * The order of elements doesn't matter (and we can't order them anyways...)
     * @param list the list of elements to compare to
     * @return wether the set version of this list, contains exactly the elements in the given list, seen as set
     */
    public boolean setEquals(KatjaList<? extends L> list) {

        // be sure we know how a set of our elements looks like
        if(set == null) toSet();

        // compare the size of both sets
        if(set.size() != list.toSet().size()) return false;

        // go through the elements of the given set and look if this set contains those
        // (this check is enough, since the set property already holds for both lists)
        for(L element : list.toSet()) if(!set.contains(element)) return false;

        // return true, as both sets have the same size, and one is included in the other
        return true;
    }

    /**
     * The replace method; lists don't have component based replaced methods, but a general one
     * @param pos the index to replace
     * @param term the term to put in place of the old
     * @return the new list, with the position replaced
     */
    public KatjaList<L> replace(int pos, Object term) {

        // if negative it's out of bounds
        if(pos < 0)
            throw new IndexOutOfBoundsException("replace on list sort invoked with negative parameter "+pos);

        // if index is too high, it's out of bounds, too
        if(pos >= values.length)
            throw new IndexOutOfBoundsException("replace on list sort invoked with too large parameter "+pos+
                    ", list has only "+values.length+" elements");

        // we create a new list, as the original list is immutable
        L[] temp = createArray(values.length);

        // add all values and replace the indexed one
        System.arraycopy(values, 0, temp, 0, pos);
        temp[pos] = (L) term;
        System.arraycopy(values, pos + 1, temp, pos + 1, values.length - (pos + 1));

        // return a new list with these values
        return createInstance(temp, false);
    }
}
