/*
    Katja
    Copyright (C) 2003-2007 see README file for authors

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package katja.common;

import katja.common.KatjaList;
import katja.common.KatjaListPos;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * User: paddy
 * Date: Jun 29, 2004
 * Time: 3:03:53 PM
 */
public abstract class KatjaListPosImpl<R extends KatjaSortPos<R>, T extends KatjaList<E>, L extends KatjaSortPos<R>, E> extends KatjaNodePosImpl<R, T> implements KatjaListPos<R, T, L, E> {

    // the array storing the elements of this position list
    protected L[] values;

    /**
     * Constructor of KatjaListPos; There is nothing to add to the super constructor.
     * @param parent reference to parent pos
     */
    public KatjaListPosImpl(KatjaSortPos<R> parent, T term, int pos) {
        super(parent, term, pos);

        values = createArray(term.size());
    }

    /**
     * Constructor of KatjaListPos as root.
     */
    public KatjaListPosImpl(T term) {
        super(term);

        values = createArray(term.size());
    }

    /**
     * Does the work of creating an L array, which can only be done in the generated classes,
     * as L is fixed to some subtype there (statically)
     * @param size the size of the array
     * @return a new array of type L
     */
    protected abstract L[] createArray(int size);


    /**
     * Returns an element of the list
     * @param i the number of the element to get
     * @return the ith element of this list
     */
    public L get(int i) {

        // preserve old parameter value
        int ith = i;

        // we allow negative addressing
        if(ith < 0) ith += values.length;

        // check if out of bounds
        if(ith < 0)
            throw new IndexOutOfBoundsException("get on list sort invoked with negative parameter "+i);

        // check if out of bounds
        if(ith >= values.length)
            throw new IndexOutOfBoundsException("get on list sort invoked with too large parameter "+i+
                    ", list has only "+values.length+" elements");

        // check if already initialized
        if(values[ith] == null)
            values[ith] = getElementInstance(ith);

        // return the value
        return values[ith];
    }

    /**
     * Returns the first element of the list, null if list is empty
     * @return the first element of the list
     */
    public L first() {

        // check if list is empty
        if(values.length == 0)
            throw new UnsupportedOperationException("first invoked on empty list");

        // check if already initialized
        if(values[0] == null)
            values[0] = getElementInstance(0);

        // return the value
        return values[0];
    }

    /**
     * Returns the last element of the list, null if list is empty
     * @return the last element of the list
     */
    public L last() {

        // check if list is empty
        if(values.length == 0)
            throw new UnsupportedOperationException("last invoked on empty list");

        // check if already initialized
        if(values[values.length - 1] == null)
            values[values.length - 1] = getElementInstance(values.length - 1);

        // return the value
        return values[values.length - 1];
    }

    /**
     * Returns a katja list of all child positions, without the last
     * @return a katja list of all child positions, without the last
     */
    public KatjaList<L> front() {
        return toList().front();
    }

    /**
     * Returns a katja list of all child positions, without the first
     * @return a katja list of all child positions, without the first
     */
    public KatjaList<L> back() {
        return toList().back();
    }

    /**
     * This method is needed to delegate the creation of an element position in a concrete case
     * @param pos the position of the element we need the position object from
     * @return the element position object
     */
    protected abstract L getElementInstance(int pos);

    /**
     * Returns the number of children this list position has
     * @return the number of children this list position has
     */
    public int size() {
        return values.length;
    }

    /**
     * Returns wether this list position has no children
     * @return wether this list position has no children
     */
    public boolean isEmpty() {
        return values.length == 0;
    }

    /**
     * Returns wether given position is a child of this position
     * @param element the child to check for
     * @return wether this position list contains given child position
     */
    public boolean contains(L element) {

        // return wether this list is the parent of given position
        return this.equals(element.parent());
    }

    /**
     * Returns wether given positions are children of this position
     * @param list the list of children to check for
     * @return wether given positions are children of this position
     */
    public boolean containsAll(KatjaList<L> list) {
        for(L element : list)
            if(!this.contains(element)) return false;

        return true;
    }

    /**
     * Returns an iterator for the lists elements
     * @return an iterator for the lists elements
     */
    public Iterator<L> iterator() {

        // return an anonymous iterator
        return new Iterator<L>() {
            int current = 0;

            public boolean hasNext() {
                return current < values.length;
            }

            public L next() {

                // check if an element left
                if(current < values.length) {

                    // check if initialized
                    if(values[current] == null)
                        values[current] = getElementInstance(current);

                    // return the element
                    return values[current++];
                }

                // if none left throw an exception
                else throw new NoSuchElementException();
            }

            public void remove() {
                throw new UnsupportedOperationException("remove invoked on list position iterator");
            }
        };
    }

    /**
     * Returns wether the term at given position is contained in the term list of this position
     * @param element the position holding the term in quesstion
     * @return wether the term at given position is contained in the term list of this position
     */
    public boolean termContains(KatjaTermPos<?, ? extends E> element) {
        return _term.contains(element.term());
    }

    /**
     * Returns wether the terms at given positions in the list are contained in the term list of this position
     * @param list the list of positions holding the terms in question
     * @return wether the terms at given positions in the list are contained in the term list of this position
     */
    public boolean termContainsAll(KatjaListPos<?, KatjaList<? extends E>, ? extends KatjaSortPos<R>, ?> list) {
        return _term.containsAll(list.term());
    }

    /**
     * A better types version of replace, which returns a list position, if the term to replace with is a list
     * of correct type.
     * @param term a term list of the same kind a this position has
     * @return a position list for given term
     */
    public KatjaListPos<R, T, L, E> replace(T term) {
        return (KatjaListPos<R, T, L, E>) super.replace(term);
    }
}
