/*
    Katja
    Copyright (C) 2003-2007 see README file for authors

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package katja.common;

import java.util.*;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.io.*;

/**
 * User: paddy
 * Date: Feb 24, 2006
 * Time: 11:26:25 AM
 */
public class KatjaSpecification {

    // the term sharing hash, using a special comparator including the "logic" when terms are equal, even if they
    // are NOT the same object, (thereby realizing term sharing)
    protected static Map<KatjaTerm, KatjaTerm> _termcache = new TreeMap<KatjaTerm, KatjaTerm>(

            // use a special comparator for the hash
            new Comparator<KatjaElement>() {

                // used to create an unique number for objects, modulo equality
                private final Map<Object, Integer> ordering = new HashMap<Object, Integer>(16384);
                private int orderingCounter = 0;

                /**
                 * Compare to simple ints
                 * @param a first int
                 * @param b second int
                 * @return wether a is smaller, equal or greater compared to b
                 */
                private int compareInts(int a, int b) {
                    if(a == b) return 0;
                    return a > b ? 1 : -1;
                }

                /**
                 * Compares two katja elements with two main goals:
                 * 1) to identify elements that could share one instance (so this is the old equals implementation before term
                 *    sharing was introduced
                 * 2) to give elements an order to put them in tree maps, etc. (the particular order is not of interest, the elements
                 *    have to be ordered fast.
                 * @param first the first element
                 * @param second the second element
                 * @return returns a stable ordering of two elements, or 0 if they are equal in the sense of katja
                 */
                public int compare(KatjaElement first, KatjaElement second) {

                    // check if one parameter is null
                    if(first == null) return -1;
                    if(second == null) return 1;

                    // check if they are trivially equal, if they are not we gained nothing; katja terms may be "equal" in our
                    // sense, cause this comparator will be used for term sharing (so before terms have the same reference)
                    if(first.equals(second)) return 0;

                    // save compare results here, to avoid double evaluation
                    int cmp;

                    // check the type
                    cmp = compareInts(System.identityHashCode(first.getClass()), System.identityHashCode(second.getClass()));
                    if(cmp != 0) return cmp;

                    // get the size
                    int size = first.size();

                    // if they don't have the same size: we are done
                    cmp = size - second.size();
                    if(cmp != 0) return cmp;

                    // so same type, same size, go to component level
                    for(int i = 0; i < size; i++) {

                        // get the components
                        Object fcomp = first.get(i);
                        Object scomp = second.get(i);

                        // we don't assume elements are equal to none elements and we assume
                        // elements to be (reference)equal, as components are already shared
                        // so if they are equal: continue
                        if(fcomp.equals(scomp)) continue;

                        // otherwise do a hash code comparison
                        cmp = compareInts(fcomp.hashCode(), scomp.hashCode());
                        if(cmp != 0) return cmp;

                        // check if ordering counter too large
                        if(orderingCounter > Integer.MAX_VALUE - 16)
                            throw new IllegalStateException("too many extern objects to compare");

                        // if they are not equal give them a number and return the comparison
                        if(!ordering.containsKey(fcomp)) ordering.put(fcomp, orderingCounter++);
                        if(!ordering.containsKey(scomp)) ordering.put(scomp, orderingCounter++);

                        // return the ordering
                        return ordering.get(fcomp) - ordering.get(scomp);
                    }

                    // so they have the same type, size and all components are equal: they are the same elements
                    return 0;
                }
            });

    /**
     * Used by the more specific sort structures. They will invoke it with their own creator implementation.
     * @param reader the input read, will be wrapped in a buffered reader, so no need to supply one
     * @param creator the creator object doing the interesting work of calling real constructors
     * @return the finished element
     * @throws IOException if something goes wrong with the reader
     */
    protected static KatjaElement parseInternal(Reader reader, ElementCreator creator) throws IOException {

        // create a buffered, line number aware reader
        LineNumberReader input = new LineNumberReader(reader);

        // get the next input line
        String line;

        // stores the current stack "bottom"
        Stack<Integer> mark = new Stack<Integer>();
        mark.push(0);

        // create a stack for our interpretation process
        Stack<Object> stack = new Stack<Object>();

        // as long as we have input: go on (an #end terminates the loop)
        while((line = input.readLine()) != null) {

            // get the line number, before descending further in the input
            int lineNumber = input.getLineNumber();

            // trim the line
            line = line.trim();
            if(line.length() == 0 || line.charAt(0) != '#')
                throw new IllegalArgumentException("command expected in line "+lineNumber+": "+line);

            // get the command
            int endOfCommand = line.indexOf(" ") < 0 ? line.length() : line.indexOf(" ");
            String command = line.substring(1, endOfCommand).toLowerCase();
            line = line.substring(endOfCommand).trim();

            // load puts an element from the beginning of the stack on the top
            if(command.equals("load")) {

                try {

                    // get the index, positive numbers address from bottom, negative from top
                    int index = Integer.valueOf(line);
                    if(index < 0) index += stack.size();
                    else index += mark.peek();

                    // check if over mark
                    if(index < mark.peek()) throw new ArrayIndexOutOfBoundsException();

                    // copy the specified stack element on top
                    stack.push(stack.get(index));
                } catch(NumberFormatException e) {
                    throw new IllegalArgumentException("argument of #load is no integer, in line "+lineNumber);
                } catch(ArrayIndexOutOfBoundsException e) {
                    throw new IllegalArgumentException("argument of #load is out of bounds, in line "+lineNumber);
                }

            // follow invokes get with all given integers on the top element
            } else if(command.equals("mark")) {

                // set a new mark for the stack
                mark.push(stack.size());

            // follow invokes get with all given integers on the top element
            } else if(command.equals("follow")) {

                // get the list of integers
                for(String item : line.split(" ")) {

                    try {

                        // replace top stack element with one of it's children
                        stack.push(((KatjaElement) stack.pop()).get(Integer.valueOf(item.trim())));
                    } catch(NumberFormatException e) {
                        throw new IllegalArgumentException("argument of #follow is no integer, in line "+lineNumber);
                    } catch(EmptyStackException e) {
                        throw new IllegalArgumentException("stack has no element to apply #follow to, in line "+lineNumber);
                    } catch(ClassCastException e) {
                        throw new IllegalArgumentException("top stack element is not an element, in line "+lineNumber);
                    }
                }

            // apply invokes the given constructor name on the top elements of the stack (number given by arity)
            // the generic algorithm needs a special creator for this task, defined in eachs specification class
            } else if(command.equals("apply")) {

                // check for two arguments
                if(line.indexOf(" ") == -1)
                    throw new IllegalArgumentException("too few arguments for #apply, in line "+lineNumber);

                // get the constructor and arity
                String constructor = line.substring(0, line.indexOf(" ")).trim();
                int arity;
                try {
                    arity = Integer.valueOf(line.substring(line.indexOf(" ")).trim());
                } catch(NumberFormatException e) {
                    throw new IllegalArgumentException("arity argument for #apply is not an Integer, in line "+lineNumber);
                }

                // check for enough elements on stack
                if(stack.size() - mark.peek() < arity)
                    throw new IllegalArgumentException((stack.size() - mark.peek())+
                            " elements are not enough for #apply, in line "+lineNumber);
                if(arity < 0)
                    throw new IllegalArgumentException("negative arity "+arity+" for #apply given, in line "+lineNumber);

                // prepare arguments for constructor
                Object[] arguments = new Object[arity];
                for(int i = 0; i < arity; i++) arguments[i] = stack.pop();

                // push the created element on the stack
                stack.push(creator.create(constructor, arguments, lineNumber));

            // ends the input or a mark
            } else if(command.equals("end")) {

                // check the top stack element
                if(stack.size() - mark.peek() < 1)
                    throw new IllegalArgumentException("no element left on stack, at end of program in line "+line);

                // clean elements up to mark
                int count = stack.size() - (mark.peek() + 1);

                // save the top element
                Object temp = stack.pop();

                // clean the elements
                for(int i = 0; i < count; i++)
                    stack.pop();

                // push saved element back up
                stack.push(temp);

                // end the current mark
                mark.pop();

                // only if all marks are done, we return the element, otherwise continue
                if(mark.isEmpty()) {

                    // check the element
                    if(!(stack.peek() instanceof KatjaElement))
                        throw new IllegalArgumentException("top stack element is no katja element, at end of program in line "+line);

                    // return it
                    return (KatjaElement) stack.peek();
                }

            // all other commands define that the following element is pushed on the stack, type defined by the command
            } else if(command.equals("string")) {

                // let's see if we have a string
                if(line.charAt(0) != '"' || line.charAt(line.length() - 1) != '"')
                    throw new IllegalArgumentException("no string given to #string, in line "+lineNumber+": "+line);

                // cut the marks off
                line = line.substring(1, line.length() - 1);

                // replace escape sequences with original content, i.e. '\\' by '\' and '\n' by a newline

                // we create the new result string in a builder, the index marks the next backslash in the original
                StringBuilder result = new StringBuilder();
                int index;

                // so if there is anything left from the input and it contains a backslash, do something
                while(line.length() > 0 && (index = line.indexOf("\\")) >= 0) {

                    // append everything before backslash to result
                    result = result.append(line.substring(0, index));

                    // look at character after backslash, if it's another one append a backslash
                    if(line.charAt(index + 1) == '\\') {
                        result.append('\\');

                    // if it's an 'n', append a newline
                    } else if(line.charAt(index + 1) == 'n') {
                        result.append('\n');

                    // everything else: the user made an error escaping the string
                    } else throw new IllegalArgumentException("Unknown escape sequence \\"+line.charAt(index + 1)+
                            " in string in line "+lineNumber);

                    // cut of the part of the string we processed
                    line = line.substring(index + 2);
                }

                // append the rest of the original string and save it
                line = result.append(line).toString();

                // push it on the stack
                stack.push(line);

            } else if(command.equals("int")) {
                try {
                    stack.push(Integer.valueOf(line));
                } catch(NumberFormatException e) {
                    throw new IllegalArgumentException(line+" is no Integer, as claimed by input line "+lineNumber);
                }
            } else if(command.equals("byte")) {
                try {
                    stack.push(Byte.valueOf(line));
                } catch(NumberFormatException e) {
                    throw new IllegalArgumentException(line+" is no Byte, as claimed by input line "+lineNumber);
                }
            } else if(command.equals("short")) {
                try {
                    stack.push(Short.valueOf(line));
                } catch(NumberFormatException e) {
                    throw new IllegalArgumentException(line+" is no Short, as claimed by input line "+lineNumber);
                }
            } else if(command.equals("long")) {
                try {
                    stack.push(Byte.valueOf(line));
                } catch(NumberFormatException e) {
                    throw new IllegalArgumentException(line+" is no Long, as claimed by input line "+lineNumber);
                }
            } else if(command.equals("float")) {
                try {
                    stack.push(Float.valueOf(line));
                } catch(NumberFormatException e) {
                    throw new IllegalArgumentException(line+" is no Float, as claimed by input line "+lineNumber);
                }
            } else if(command.equals("double")) {
                try {
                    stack.push(Double.valueOf(line));
                } catch(NumberFormatException e) {
                    throw new IllegalArgumentException(line+" is no Double, as claimed by input line "+lineNumber);
                }
            } else if(command.equals("boolean")) {
                try {
                    stack.push(Boolean.valueOf(line));
                } catch(NumberFormatException e) {
                    throw new IllegalArgumentException(line+" is no Boolean, as claimed by input line "+lineNumber);
                }
            } else throw new IllegalArgumentException("unknown command "+command+", in line "+lineNumber);
        }

        // we should not leave the loop, only if the input ended but we were not finished
        throw new IllegalArgumentException("unexpected end of input");
    }

    /**
     * Specification classes have to implement at least one creator, if they want to use the generic parse method
     */
    public static interface ElementCreator {

        /**
         * Creates an katja element of given sort, with given parameters. The creator has to check the parameter types
         * and if such a sort exists at all. It also has to check the arity.
         * @param constructor the name of the sort or constructor
         * @param parameters the parameters, given as objects
         * @param line the line number where the command appeared, used for error messages
         * @return the constructed element, if possible
         */
        public KatjaElement create(String constructor, Object[] parameters, int line);
    }
}
