/*
    Katja
    Copyright (C) 2003-2007 see README file for authors

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package katja.common;

import java.util.*;

public abstract class KatjaTermImpl extends KatjaElementImpl implements KatjaTerm {

    // used to create an unique number for uncomparable objects
    private static final Map<Object, Integer> ordering = new HashMap<Object, Integer>();
    private static int orderingCounter = 1;
    
    /**
     * Create an iterator over all subterms of a term. If leaf terms are external katja elements, this iterator
     * will descend to those elements.
     * @return an iterator over all subterms, in unspecified order
     */
    public Iterator<Object> iterate() {

        // create a queue to store the elements, which are not yet returned
        final Queue<Object> q = new LinkedList<Object>();

        // add the start element, which is this
        q.add(this);

        // return the iterator doing the job
        return new Iterator<Object>() {

            // as long as the queue is not empty we can return an unseen element
            public boolean hasNext() {
                return q.peek() != null;
            }

            // the next element is simply the queues head, but we need to take care of all children, if there are any
            public Object next() {

                // get the first element
                Object element = q.poll();

                // if there is the possiblity of children
                if(element instanceof KatjaElement)

                // go through the children
                for(int i = ((KatjaElement) element).size() - 1; i >= 0; i-- ) {

                    // add the child to the queue
                    q.add(((KatjaElement) element).get(i));
                }

                // return the head element
                return element;
            }

            // disable remove, as this iterator is for visits only
            public void remove() {
                throw new UnsupportedOperationException("remove is not supported on KatjaTerm iterators");
            }
        };
    }

    /**
     * Compare this object with another KatjaTerm {@link java.lang.Comparable#compareTo}. We return 1 if
     * {@code katjaSort} is null. If {@code this} and {@code katjaSort} are equal, we return 0. Then both KatjaSorts are
     * compared. If they are equal, the childs of {@code this} are compared with the childs of {@code katjaSort} from
     * left to right lexikographically combined.
     *
     * The comparison of childs which are instances of {@code KatjaTerm} itself is delegated to the {@code compareTo}
     * method of the child of {@code this}.
     * The comparison of childs which are not instances of {@code KatjaTerm} but instances of {@code Comparable} is
     * delegated to the {@code compareTo} method of the child of {@code this}. If this fails with an ClassCastException
     * it is delegated to the {@code compareTo} method of the child of {@code katjaSort}.
     * If we still could not compare {@code this} and {@code katjaSort} we lexikographically combine the orderings of
     * the results of {@code getClass().getName()}, of {@code hashCode()} and an arbitrary but fixed one.
     *
     * To make it short: This orders every {@code KatjaTerm} in a deterministic way. You only have to worry about non
     * Katja generated classes which are not Comparable. And only if the hashcode of those objects is not the memory
     * address of the object itself. This is the case with most JDKs. {@see java.lang.Object#hashCode()}.
     * @param katjaSort
     * @return -1,0,+1
     */
    public int compareTo(KatjaSort katjaSort)
    {
        // if compared to null we just say 1
        if(katjaSort == null) return 1;

        // if they are equal return 0
        if(this.equals(katjaSort)) return 0;

        // compare the sort names; if they don't match we have our ordering
        int cmp = this.getClass().getSimpleName().compareTo(katjaSort.getClass().getSimpleName());
        if(cmp != 0) return cmp;

        // get the sizes of both terms
        int sizeA = this.size();
        int sizeB = katjaSort.size();

        // go through the components
        for(int i = 0; i < sizeA && i < sizeB; i++) {

            // fst and snd may not ne null
            final Object fst = this.get(i);
            final Object snd = katjaSort.get(i);

            // if a way to compare the components has been found
            boolean found = false;

            // if both are terms just compare them
            if(fst instanceof KatjaSort && snd instanceof KatjaSort) {
                cmp = ((KatjaSort) fst).compareTo(((KatjaSort) snd));
                found = true;

            // if both are comparable (compareTo has to be symmetric) try to compare them
            } else if(fst instanceof Comparable && snd instanceof Comparable) {
                try {
                    cmp = ((Comparable) fst).compareTo((Comparable) snd);
                    found = true;
                } catch (ClassCastException ce) {
                    try {
                        cmp = - ((Comparable) snd).compareTo((Comparable) fst);
                        found = true;
                    } catch(ClassCastException ce2) {
                    }
                }
            }

            // if none of the above methods can be used to compare the components we have to try something else
            if(!found) {

                // let's compare the classname
                cmp = fst.getClass().getName().compareTo(snd.getClass().getName());

                // if they are of the same name we still have to try something else
                if (cmp == 0) {

                    // let's compare the hashcode
                    cmp = new Integer(System.identityHashCode(fst)).compareTo(System.identityHashCode(snd));

                    // if they are the same too, but the objects are not indicated as equal, we have to seperate them
                    if(cmp == 0 && !fst.equals(snd)) {

                        // we give them both a unique number
                        if (ordering.get(fst) == null) {
                            ordering.put(fst, orderingCounter);
                            orderingCounter += 1;
                        }
                        if (ordering.get(snd) == null) {
                            ordering.put(snd, orderingCounter);
                            orderingCounter += 1;
                        }

                        // and use these numbers to get an ordering
                        cmp = ordering.get(fst).compareTo(ordering.get(snd));
                    }
                }
            }

            // if we have successfully ordered the elements we can break the loop
            if (cmp != 0) break;
        }

        // if they are still equal we compare the size of both terms
        if(cmp == 0) cmp = new Integer(sizeA).compareTo(sizeB);

        // at this point we require the terms to be different
        //assert cmp != 0;

        // return the result
        return cmp;
    }
}
