/*
    Katja
    Copyright (C) 2003-2007 see README file for authors

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package katja.common;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * User: paddy
 * Date: Jun 29, 2004
 * Time: 2:27:20 PM
 */
public abstract class KatjaTermPosImpl<R extends KatjaSortPos<R>, T extends Object> extends KatjaElementImpl implements KatjaTermPos<R, T> {

    // reference to the upper context of the pos.
    protected KatjaSortPos<R> _parent;

    // reference to the root position of this position structure.
    protected R _root;

    // reference to the term of this position.
    protected T _term;

    // indicates which child this position is, seen from it's parents view. -1 if root.
    protected int _position;

    /**
     * KatjaTermPos constructor takes the upper context of the position.
     * @param parent reference to parent pos
     */
    protected KatjaTermPosImpl(KatjaSortPos<R> parent, T term, int pos) {

        // initialize essential attributes
        this._parent = parent;
        this._term = term;
        this._position = pos;

        // look for parent to initialize root
        if(parent != null) {
            this._root = parent.root();

        // if we have no parent, but are not position -1, something went wrong...
        } else if(pos != -1) {
            throw new IllegalArgumentException("term position constructor called with no parent but position "+pos);

        // so this is the case of the position being a root
        } else {

            // we don't know what type R is, but we assume it is the sort of this, everything else would be an error
            this._root = (R) this;
        }
    }

    /**
     * Returns the position of this pos relative to it's parent
     * @return the position of this pos relative to it's parent
     */
    public int position() {
        return _position;
    }

    /**
     * Getter method for the root reference.
     * @return the root reference for this structure, stored in this position
     */
    public R root() {
        return _root;
    }

    /**
     * Getter method for the term reference.
     * @return the term reference to this position
     */
    public T term() {
        return _term;
    }

    /**
     * Getter method for the parent reference.
     * @return the parent pos of this pos
     */
    public KatjaSortPos<R> parent() {
        return _parent;
    }

    /**
     * Returns the right sibling of this position
     * @return the right sibling of this position
     */
    public KatjaSortPos<R> rsib() {

        // if we don't have a parent, we don't have right siblings
        if(parent() == null) return null;

        // if our position is not the last, get the next sibling
        return position() < parent().size() - 1 ? parent().get(position() + 1) : null;
    }

    /**
     * Returns the left sibling of this position
     * @return the left sibling of this position
     */
    public KatjaSortPos<R> lsib() {

        // if we don't have a parent, we don't have left siblings
        if(parent() == null) return null;

        // if our position is not the first, get the next sibling
        return position() > 0 ? parent().get(position() - 1) : null;
    }

    /**
     * Walks through the positions in pre-order
     * @return the position next to this in pre-order
     */
    public KatjaSortPos<R> preOrder() {

        // descend to children
        if(size() > 0) return get(0);

        return preOrderSkip();
    }

    public KatjaSortPos<R> preOrderSkip() {
        // no children? if no parent return null;
        if(parent() == null) return null;

        // no children but parent? goto right sibling
        if(position() < parent().size() - 1) return parent().get(position() + 1);

        // no right sibling? find a parent who has
        KatjaSortPos<R> parent = parent();

        while(parent != null && parent.rsib() == null) parent = parent.parent();

        // return the sibling, null if finished
        if(parent == null) return null;
        else return parent.rsib();
    }

    /**
     * Returns the first position to visit when threading in preorder
     * @return the first position to visit when threading in preorder
     */
    public KatjaSortPos<R> postOrderStart() {

        // start with this
        KatjaSortPos<R> pos = this;

        // descend depth first till we get to a position without children
        while(pos.size() > 0) pos = pos.get(0);

        // return this position
        return pos;
    }

    /**
     * Walks through the positions in post-order
     * @return the position next to this in post-order
     */
    public KatjaSortPos<R> postOrder() {
        if(parent() == null) return null;

        // if there are no right siblings return to parent
        if(position() >= parent().size() - 1) return parent();

        // so if there are siblings go to the next and descend
        return parent().get(position() + 1).postOrderStart();
    }

    /**
     * Tests the two associated terms of positions for equality.
     * @param pos the position to compare with
     * @return whether the term of this position equals the term of the parameter position
     */
    public boolean termEquals(KatjaSortPos<?> pos) {
        return term().equals(pos.term());
    }

    /**
     * Returns the path from the root to this position, as integer list which can be used in follow
     * @return the path from the root to this position
     */
    public List<Integer> path() {

        // start with this
        KatjaSortPos<R> pos = this;

        // save the position of the current node in a list
        List<Integer> list = new ArrayList<Integer>();

        // go back to the root
        while(pos.parent() != null) {
            list.add(pos.position());
            pos = pos.parent();
        }

        // we have the path from the position to the root, but we want it the othwer way round
        Collections.reverse(list);

        // return the result
        return list;
    }

    /**
     * Returns the path from given position to this position, as integer list which can be used in follow.
     * This method does NOT require bot positions to be in the same structure, it does however require the resulting
     * position (after merging pathes) to have the same underlying term. This may not necessarily mean you get
     * something meaningfull from this method. This is done so you can compare positions from structure parts, which
     * did not change on operations.
     * @param start the start position to use
     * @return the path from given position to this position
     */
    public List<Integer> pathFrom(KatjaSortPos<R> start) {

        // get both pathes from root
        List<Integer> pathToSource = start.path();
        List<Integer> pathToTarget = path();

        // cut off the common part
        while(!pathToSource.isEmpty() && !pathToTarget.isEmpty()) {
            if(pathToSource.get(0) == pathToTarget.get(0)) {
                pathToSource.remove(0);
                pathToTarget.remove(0);
            } else break;
        }

        // now construct the result list
        List<Integer> result = new ArrayList<Integer>();

        // go from source to first common parent node
        for(int i = 0; i < pathToSource.size(); i++) result.add(-1);

        // and then down to this
        result.addAll(pathToTarget);

        // now test this path
        try {
            if(!start.follow(result).termEquals(this)) throw new IllegalArgumentException();
        } catch(IllegalArgumentException e) {
            throw new IllegalArgumentException("pathFrom got a start position, which is not compatible with this");
        }

        // return this list
        return result;
    }

    /**
     * Follows a given path in the position structure, returning the described position
     * @param path the path to follow as numbers of children or -1 for parent
     * @return the position situated relative to this one if one follows the path
     */
    public KatjaSortPos<R> follow(List<Integer> path) {

        // start at the current position
        KatjaSortPos<R> termPos = this;

        // go through the given list
        for(Integer i : path) {

            // -1 specifies ther parent, this can be null, if current term was root!
            if(i == -1) {

                // go to the parent
                termPos = termPos.parent();

                // so if it is null: throw an exception
                if(termPos == null)
                    throw new IllegalArgumentException("follow got a path which exits the position structure at the root position");

            // otherwise the nunmber specifies a child
            } else {

                // if the number is out of bounds: throw an exception
                if(i < 0 || i >= termPos.size())
                    throw new IllegalArgumentException("follow got a path selecting non-existing child "+i+" (of "+termPos.size()+")");

                // if not, descend to the child
                termPos = termPos.get(i);
            }
        }

        // return the result position
        return termPos;
    }

    /**
     * Substitutes the term at the current position with the given term.
     * This operations parameter can not be typed much better than Object, as we allow to replace leaves too.
     * To guarantee the return-type, the term has to be checked against the allow root sorts.
     * The operation returns the new position after substitution.
     * @param term the term to put at the current position.
     * @return a new position representing the new term in the same context as before
     */
    public KatjaSortPos<R> replace(Object term) {

        // check if the position is a root position
        if(parent() == null) {

            if(!(term instanceof KatjaSort))
                throw new IllegalArgumentException("given term to replace root position has some extern sort");

            // just return the new position; type checking has to be done in the fresh method,
            // but the given term has to be at least a katja sort, as roots can't be extern sorts
            return freshRootPosition((KatjaSort) term);
        }

        // get the current position
        KatjaSortPos<R> pos = this;

        // now get the more general super terms, up to the root
        while(pos.parent() != null) {
            term = ((KatjaNodePos<R, ? extends KatjaTerm>) pos.parent()).term().replace(pos.position(), term);
            pos = pos.parent();
        }

        // get a new position for the super term, we know term has to be a katja sort by now
        KatjaSortPos<R> result = freshRootPosition((KatjaSort) term);

        // follow the path back to the old 'this' position
        result = result.follow(path());

        // return the position
        return result;
    }

    /**
     * Returns the root position to given term. Can only be created if known in which structure. The concrete type
     * shall delegate it to the factory.
     * @param term the term to get the root position from
     * @return the root position to given term
     */
    protected abstract KatjaSortPos<R> freshRootPosition(KatjaSort term);

    /**
     * Lexicographic combination of {@code this.root().term().compareTo(katjaSortPos.root().term())} and the comparison of the
     * elements of the result of {@code this.path()} and {@code katjaSortPos.path()}
     * @param katjaSortPos
     * @return -1,0,+1
     */
    public int compareTo(KatjaSortPos katjaSortPos) {

        int cmp = ((KatjaNodePos<R, ? extends KatjaTerm>) this.root()).term().
                compareTo(((KatjaNodePos<R, ? extends KatjaTerm>) katjaSortPos.root()).term());

        if (cmp == 0) {
            List<Integer> pathA = this.path();
            List<Integer> pathB = katjaSortPos.path();
            int lengthA = pathA.size();
            int lengthB = pathB.size();

            for (int i=0; i<lengthA && i < lengthB; i++) {
                cmp = pathA.get(i).compareTo(pathB.get(i));
                if (cmp != 0) break;
            }
            if (cmp == 0) {
                cmp = new Integer(lengthA).compareTo(lengthB);
            }
        }

        return cmp;
    }
}
